# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: SyncDown.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import record_pb2 as record__pb2
from . import breachwatch_pb2 as breachwatch__pb2
from . import APIRequest_pb2 as APIRequest__pb2
from . import enterprise_pb2 as enterprise__pb2
from . import NotificationCenter_pb2 as NotificationCenter__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0eSyncDown.proto\x12\x05Vault\x1a\x0crecord.proto\x1a\x11\x62reachwatch.proto\x1a\x10\x41PIRequest.proto\x1a\x10\x65nterprise.proto\x1a\x18NotificationCenter.proto\"A\n\x0fSyncDownRequest\x12\x19\n\x11\x63ontinuationToken\x18\x01 \x01(\x0c\x12\x13\n\x0b\x64\x61taVersion\x18\x02 \x01(\x05\"\x81\x11\n\x10SyncDownResponse\x12\x19\n\x11\x63ontinuationToken\x18\x01 \x01(\x0c\x12\x0f\n\x07hasMore\x18\x02 \x01(\x08\x12\'\n\x0b\x63\x61\x63heStatus\x18\x03 \x01(\x0e\x32\x12.Vault.CacheStatus\x12&\n\x0buserFolders\x18\x04 \x03(\x0b\x32\x11.Vault.UserFolder\x12*\n\rsharedFolders\x18\x05 \x03(\x0b\x32\x13.Vault.SharedFolder\x12>\n\x17userFolderSharedFolders\x18\x06 \x03(\x0b\x32\x1d.Vault.UserFolderSharedFolder\x12\x36\n\x13sharedFolderFolders\x18\x07 \x03(\x0b\x32\x19.Vault.SharedFolderFolder\x12\x1e\n\x07records\x18\x08 \x03(\x0b\x32\r.Vault.Record\x12-\n\x0erecordMetaData\x18\t \x03(\x0b\x32\x15.Vault.RecordMetaData\x12+\n\rnonSharedData\x18\n \x03(\x0b\x32\x14.Vault.NonSharedData\x12&\n\x0brecordLinks\x18\x0b \x03(\x0b\x32\x11.Vault.RecordLink\x12\x32\n\x11userFolderRecords\x18\x0c \x03(\x0b\x32\x17.Vault.UserFolderRecord\x12\x36\n\x13sharedFolderRecords\x18\r \x03(\x0b\x32\x19.Vault.SharedFolderRecord\x12\x42\n\x19sharedFolderFolderRecords\x18\x0e \x03(\x0b\x32\x1f.Vault.SharedFolderFolderRecord\x12\x32\n\x11sharedFolderUsers\x18\x0f \x03(\x0b\x32\x17.Vault.SharedFolderUser\x12\x32\n\x11sharedFolderTeams\x18\x10 \x03(\x0b\x32\x17.Vault.SharedFolderTeam\x12\x1a\n\x12recordAddAuditData\x18\x11 \x03(\x0c\x12\x1a\n\x05teams\x18\x12 \x03(\x0b\x32\x0b.Vault.Team\x12,\n\x0esharingChanges\x18\x13 \x03(\x0b\x32\x14.Vault.SharingChange\x12\x1f\n\x07profile\x18\x14 \x01(\x0b\x32\x0e.Vault.Profile\x12%\n\nprofilePic\x18\x15 \x01(\x0b\x32\x11.Vault.ProfilePic\x12\x34\n\x12pendingTeamMembers\x18\x16 \x03(\x0b\x32\x18.Vault.PendingTeamMember\x12\x34\n\x12\x62reachWatchRecords\x18\x17 \x03(\x0b\x32\x18.Vault.BreachWatchRecord\x12\"\n\tuserAuths\x18\x18 \x03(\x0b\x32\x0f.Vault.UserAuth\x12?\n\x17\x62reachWatchSecurityData\x18\x19 \x03(\x0b\x32\x1e.Vault.BreachWatchSecurityData\x12/\n\x0freusedPasswords\x18\x1a \x01(\x0b\x32\x16.Vault.ReusedPasswords\x12\x1a\n\x12removedUserFolders\x18\x1b \x03(\x0c\x12\x1c\n\x14removedSharedFolders\x18\x1c \x03(\x0c\x12\x45\n\x1eremovedUserFolderSharedFolders\x18\x1d \x03(\x0b\x32\x1d.Vault.UserFolderSharedFolder\x12=\n\x1aremovedSharedFolderFolders\x18\x1e \x03(\x0b\x32\x19.Vault.SharedFolderFolder\x12\x16\n\x0eremovedRecords\x18\x1f \x03(\x0c\x12-\n\x12removedRecordLinks\x18  \x03(\x0b\x32\x11.Vault.RecordLink\x12\x39\n\x18removedUserFolderRecords\x18! \x03(\x0b\x32\x17.Vault.UserFolderRecord\x12=\n\x1aremovedSharedFolderRecords\x18\" \x03(\x0b\x32\x19.Vault.SharedFolderRecord\x12I\n removedSharedFolderFolderRecords\x18# \x03(\x0b\x32\x1f.Vault.SharedFolderFolderRecord\x12\x39\n\x18removedSharedFolderUsers\x18$ \x03(\x0b\x32\x17.Vault.SharedFolderUser\x12\x39\n\x18removedSharedFolderTeams\x18% \x03(\x0b\x32\x17.Vault.SharedFolderTeam\x12\x14\n\x0cremovedTeams\x18& \x03(\x0c\x12&\n\x0cksmAppShares\x18\' \x03(\x0b\x32\x10.Vault.KsmChange\x12\'\n\rksmAppClients\x18( \x03(\x0b\x32\x10.Vault.KsmChange\x12\x30\n\x10shareInvitations\x18) \x03(\x0b\x32\x16.Vault.ShareInvitation\x12+\n\x0b\x64iagnostics\x18* \x01(\x0b\x32\x16.Vault.SyncDiagnostics\x12.\n\x0frecordRotations\x18+ \x03(\x0b\x32\x15.Vault.RecordRotation\x12\x1a\n\x05users\x18, \x03(\x0b\x32\x0b.Vault.User\x12\x14\n\x0cremovedUsers\x18- \x03(\x0c\x12\x33\n\x11securityScoreData\x18. \x03(\x0b\x32\x18.Vault.SecurityScoreData\x12\x41\n\x10notificationSync\x18/ \x03(\x0b\x32\'.NotificationCenter.NotificationWrapper\"\x92\x01\n\nUserFolder\x12\x11\n\tfolderUid\x18\x01 \x01(\x0c\x12\x11\n\tparentUid\x18\x02 \x01(\x0c\x12\x15\n\ruserFolderKey\x18\x03 \x01(\x0c\x12\'\n\x07keyType\x18\x04 \x01(\x0e\x32\x16.Records.RecordKeyType\x12\x10\n\x08revision\x18\x05 \x01(\x03\x12\x0c\n\x04\x64\x61ta\x18\x06 \x01(\x0c\"\xd5\x02\n\x0cSharedFolder\x12\x17\n\x0fsharedFolderUid\x18\x01 \x01(\x0c\x12\x10\n\x08revision\x18\x02 \x01(\x03\x12\x17\n\x0fsharedFolderKey\x18\x03 \x01(\x0c\x12\'\n\x07keyType\x18\x04 \x01(\x0e\x32\x16.Records.RecordKeyType\x12\x0c\n\x04\x64\x61ta\x18\x05 \x01(\x0c\x12\x1c\n\x14\x64\x65\x66\x61ultManageRecords\x18\x06 \x01(\x08\x12\x1a\n\x12\x64\x65\x66\x61ultManageUsers\x18\x07 \x01(\x08\x12\x16\n\x0e\x64\x65\x66\x61ultCanEdit\x18\x08 \x01(\x08\x12\x19\n\x11\x64\x65\x66\x61ultCanReshare\x18\t \x01(\x08\x12\'\n\x0b\x63\x61\x63heStatus\x18\n \x01(\x0e\x32\x12.Vault.CacheStatus\x12\r\n\x05owner\x18\x0b \x01(\t\x12\x17\n\x0fownerAccountUid\x18\x0c \x01(\x0c\x12\x0c\n\x04name\x18\r \x01(\x0c\"V\n\x16UserFolderSharedFolder\x12\x11\n\tfolderUid\x18\x01 \x01(\x0c\x12\x17\n\x0fsharedFolderUid\x18\x02 \x01(\x0c\x12\x10\n\x08revision\x18\x03 \x01(\x03\"\xbb\x01\n\x12SharedFolderFolder\x12\x17\n\x0fsharedFolderUid\x18\x01 \x01(\x0c\x12\x11\n\tfolderUid\x18\x02 \x01(\x0c\x12\x11\n\tparentUid\x18\x03 \x01(\x0c\x12\x1d\n\x15sharedFolderFolderKey\x18\x04 \x01(\x0c\x12\'\n\x07keyType\x18\x05 \x01(\x0e\x32\x16.Records.RecordKeyType\x12\x10\n\x08revision\x18\x06 \x01(\x03\x12\x0c\n\x04\x64\x61ta\x18\x07 \x01(\x0c\"l\n\x0fSharedFolderKey\x12\x17\n\x0fsharedFolderUid\x18\x01 \x01(\x0c\x12\x17\n\x0fsharedFolderKey\x18\x02 \x01(\x0c\x12\'\n\x07keyType\x18\x03 \x01(\x0e\x32\x16.Records.RecordKeyType\"\xc3\x02\n\x04Team\x12\x0f\n\x07teamUid\x18\x01 \x01(\x0c\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x0f\n\x07teamKey\x18\x03 \x01(\x0c\x12+\n\x0bteamKeyType\x18\x04 \x01(\x0e\x32\x16.Records.RecordKeyType\x12\x16\n\x0eteamPrivateKey\x18\x05 \x01(\x0c\x12\x14\n\x0crestrictEdit\x18\x06 \x01(\x08\x12\x15\n\rrestrictShare\x18\x07 \x01(\x08\x12\x14\n\x0crestrictView\x18\x08 \x01(\x08\x12\x1c\n\x14removedSharedFolders\x18\t \x03(\x0c\x12\x30\n\x10sharedFolderKeys\x18\n \x03(\x0b\x32\x16.Vault.SharedFolderKey\x12\x19\n\x11teamEccPrivateKey\x18\x0b \x01(\x0c\x12\x18\n\x10teamEccPublicKey\x18\x0c \x01(\x0c\"\xbf\x01\n\x06Record\x12\x11\n\trecordUid\x18\x01 \x01(\x0c\x12\x10\n\x08revision\x18\x02 \x01(\x03\x12\x0f\n\x07version\x18\x03 \x01(\x05\x12\x0e\n\x06shared\x18\x04 \x01(\x08\x12\x1a\n\x12\x63lientModifiedTime\x18\x05 \x01(\x03\x12\x0c\n\x04\x64\x61ta\x18\x06 \x01(\x0c\x12\r\n\x05\x65xtra\x18\x07 \x01(\x0c\x12\r\n\x05udata\x18\x08 \x01(\t\x12\x10\n\x08\x66ileSize\x18\t \x01(\x03\x12\x15\n\rthumbnailSize\x18\n \x01(\x03\"b\n\nRecordLink\x12\x17\n\x0fparentRecordUid\x18\x01 \x01(\x0c\x12\x16\n\x0e\x63hildRecordUid\x18\x02 \x01(\x0c\x12\x11\n\trecordKey\x18\x03 \x01(\x0c\x12\x10\n\x08revision\x18\x04 \x01(\x03\"J\n\x10UserFolderRecord\x12\x11\n\tfolderUid\x18\x01 \x01(\x0c\x12\x11\n\trecordUid\x18\x02 \x01(\x0c\x12\x10\n\x08revision\x18\x03 \x01(\x03\"k\n\x18SharedFolderFolderRecord\x12\x17\n\x0fsharedFolderUid\x18\x01 \x01(\x0c\x12\x11\n\tfolderUid\x18\x02 \x01(\x0c\x12\x11\n\trecordUid\x18\x03 \x01(\x0c\x12\x10\n\x08revision\x18\x04 \x01(\x03\"0\n\rNonSharedData\x12\x11\n\trecordUid\x18\x01 \x01(\x0c\x12\x0c\n\x04\x64\x61ta\x18\x02 \x01(\x0c\"\x9f\x02\n\x0eRecordMetaData\x12\x11\n\trecordUid\x18\x01 \x01(\x0c\x12\r\n\x05owner\x18\x02 \x01(\x08\x12\x11\n\trecordKey\x18\x03 \x01(\x0c\x12-\n\rrecordKeyType\x18\x04 \x01(\x0e\x32\x16.Records.RecordKeyType\x12\x10\n\x08\x63\x61nShare\x18\x05 \x01(\x08\x12\x0f\n\x07\x63\x61nEdit\x18\x06 \x01(\x08\x12\x17\n\x0fownerAccountUid\x18\x07 \x01(\x0c\x12\x12\n\nexpiration\x18\x08 \x01(\x03\x12\x42\n\x1a\x65xpirationNotificationType\x18\t \x01(\x0e\x32\x1e.Records.TimerNotificationType\x12\x15\n\rownerUsername\x18\n \x01(\t\"2\n\rSharingChange\x12\x11\n\trecordUid\x18\x01 \x01(\x0c\x12\x0e\n\x06shared\x18\x02 \x01(\x08\">\n\x07Profile\x12\x0c\n\x04\x64\x61ta\x18\x01 \x01(\x0c\x12\x13\n\x0bprofileName\x18\x02 \x01(\t\x12\x10\n\x08revision\x18\x03 \x01(\x03\"+\n\nProfilePic\x12\x0b\n\x03url\x18\x01 \x01(\t\x12\x10\n\x08revision\x18\x02 \x01(\x03\"p\n\x11PendingTeamMember\x12\x18\n\x10\x65nterpriseUserId\x18\x01 \x01(\x03\x12\x15\n\ruserPublicKey\x18\x02 \x01(\x0c\x12\x10\n\x08teamUids\x18\x03 \x03(\x0c\x12\x18\n\x10userEccPublicKey\x18\x04 \x01(\x0c\"\xa6\x01\n\x11\x42reachWatchRecord\x12\x11\n\trecordUid\x18\x01 \x01(\x0c\x12\x0c\n\x04\x64\x61ta\x18\x02 \x01(\x0c\x12.\n\x04type\x18\x03 \x01(\x0e\x32 .BreachWatch.BreachWatchInfoType\x12\x11\n\tscannedBy\x18\x04 \x01(\t\x12\x10\n\x08revision\x18\x05 \x01(\x03\x12\x1b\n\x13scannedByAccountUid\x18\x06 \x01(\x0c\"\xb4\x01\n\x08UserAuth\x12\x0b\n\x03uid\x18\x01 \x01(\x0c\x12,\n\tloginType\x18\x02 \x01(\x0e\x32\x19.Authentication.LoginType\x12\x0f\n\x07\x64\x65leted\x18\x03 \x01(\x08\x12\x12\n\niterations\x18\x04 \x01(\x05\x12\x0c\n\x04salt\x18\x05 \x01(\x0c\x12\x1a\n\x12\x65ncryptedClientKey\x18\x06 \x01(\x0c\x12\x10\n\x08revision\x18\x07 \x01(\x03\x12\x0c\n\x04name\x18\x08 \x01(\t\">\n\x17\x42reachWatchSecurityData\x12\x11\n\trecordUid\x18\x01 \x01(\x0c\x12\x10\n\x08revision\x18\x02 \x01(\x03\"2\n\x0fReusedPasswords\x12\r\n\x05\x63ount\x18\x01 \x01(\x05\x12\x10\n\x08revision\x18\x02 \x01(\x03\"\xa9\x02\n\x12SharedFolderRecord\x12\x17\n\x0fsharedFolderUid\x18\x01 \x01(\x0c\x12\x11\n\trecordUid\x18\x02 \x01(\x0c\x12\x11\n\trecordKey\x18\x03 \x01(\x0c\x12\x10\n\x08\x63\x61nShare\x18\x04 \x01(\x08\x12\x0f\n\x07\x63\x61nEdit\x18\x05 \x01(\x08\x12\x17\n\x0fownerAccountUid\x18\x06 \x01(\x0c\x12\x12\n\nexpiration\x18\x07 \x01(\x03\x12\r\n\x05owner\x18\x08 \x01(\x08\x12\x42\n\x1a\x65xpirationNotificationType\x18\t \x01(\x0e\x32\x1e.Records.TimerNotificationType\x12\x15\n\rownerUsername\x18\n \x01(\t\x12\x1a\n\x12rotateOnExpiration\x18\x0b \x01(\x08\"\xf1\x01\n\x10SharedFolderUser\x12\x17\n\x0fsharedFolderUid\x18\x01 \x01(\x0c\x12\x10\n\x08username\x18\x02 \x01(\t\x12\x15\n\rmanageRecords\x18\x03 \x01(\x08\x12\x13\n\x0bmanageUsers\x18\x04 \x01(\x08\x12\x12\n\naccountUid\x18\x05 \x01(\x0c\x12\x12\n\nexpiration\x18\x06 \x01(\x03\x12\x42\n\x1a\x65xpirationNotificationType\x18\x07 \x01(\x0e\x32\x1e.Records.TimerNotificationType\x12\x1a\n\x12rotateOnExpiration\x18\x08 \x01(\x08\"\xea\x01\n\x10SharedFolderTeam\x12\x17\n\x0fsharedFolderUid\x18\x01 \x01(\x0c\x12\x0f\n\x07teamUid\x18\x02 \x01(\x0c\x12\x0c\n\x04name\x18\x03 \x01(\t\x12\x15\n\rmanageRecords\x18\x04 \x01(\x08\x12\x13\n\x0bmanageUsers\x18\x05 \x01(\x08\x12\x12\n\nexpiration\x18\x06 \x01(\x03\x12\x42\n\x1a\x65xpirationNotificationType\x18\x07 \x01(\x0e\x32\x1e.Records.TimerNotificationType\x12\x1a\n\x12rotateOnExpiration\x18\x08 \x01(\x08\"\x8a\x01\n\tKsmChange\x12\x14\n\x0c\x61ppRecordUid\x18\x01 \x01(\x0c\x12\x10\n\x08\x64\x65tailId\x18\x02 \x01(\x0c\x12\x0f\n\x07removed\x18\x03 \x01(\x08\x12\x30\n\rappClientType\x18\x04 \x01(\x0e\x32\x19.Enterprise.AppClientType\x12\x12\n\nexpiration\x18\x05 \x01(\x03\"#\n\x0fShareInvitation\x12\x10\n\x08username\x18\x01 \x01(\t\",\n\x04User\x12\x12\n\naccountUid\x18\x01 \x01(\x0c\x12\x10\n\x08username\x18\x02 \x01(\t\"{\n\x0fSyncDiagnostics\x12\x19\n\x11\x63ontinuationToken\x18\x01 \x01(\x0c\x12\x0e\n\x06userId\x18\x02 \x01(\x05\x12\x18\n\x10\x65nterpriseUserId\x18\x03 \x01(\x03\x12\x10\n\x08syncedTo\x18\x04 \x01(\x03\x12\x11\n\tsyncingTo\x18\x05 \x01(\x03\"\xee\x01\n\x0eRecordRotation\x12\x11\n\trecordUid\x18\x01 \x01(\x0c\x12\x10\n\x08revision\x18\x02 \x01(\x03\x12\x18\n\x10\x63onfigurationUid\x18\x03 \x01(\x0c\x12\x10\n\x08schedule\x18\x04 \x01(\t\x12\x15\n\rpwdComplexity\x18\x05 \x01(\x0c\x12\x10\n\x08\x64isabled\x18\x06 \x01(\x08\x12\x13\n\x0bresourceUid\x18\x07 \x01(\x0c\x12\x14\n\x0clastRotation\x18\x08 \x01(\x03\x12\x37\n\x12lastRotationStatus\x18\t \x01(\x0e\x32\x1b.Vault.RecordRotationStatus\"F\n\x11SecurityScoreData\x12\x11\n\trecordUid\x18\x01 \x01(\x0c\x12\x0c\n\x04\x64\x61ta\x18\x02 \x01(\x0c\x12\x10\n\x08revision\x18\x03 \x01(\x03\"3\n\x1d\x42reachWatchGetSyncDataRequest\x12\x12\n\nrecordUids\x18\x01 \x03(\x0c\"\xb3\x01\n\x1e\x42reachWatchGetSyncDataResponse\x12\x34\n\x12\x62reachWatchRecords\x18\x01 \x03(\x0b\x32\x18.Vault.BreachWatchRecord\x12?\n\x17\x62reachWatchSecurityData\x18\x02 \x03(\x0b\x32\x1e.Vault.BreachWatchSecurityData\x12\x1a\n\x05users\x18\x03 \x03(\x0b\x32\x0b.Vault.User\"6\n\x18GetAccountUidMapResponse\x12\x1a\n\x05users\x18\x01 \x03(\x0b\x32\x0b.Vault.User*\"\n\x0b\x43\x61\x63heStatus\x12\x08\n\x04KEEP\x10\x00\x12\t\n\x05\x43LEAR\x10\x01*f\n\x14RecordRotationStatus\x12\x14\n\x10RRST_NOT_ROTATED\x10\x00\x12\x14\n\x10RRST_IN_PROGRESS\x10\x01\x12\x10\n\x0cRRST_SUCCESS\x10\x02\x12\x10\n\x0cRRST_FAILURE\x10\x03\x42!\n\x18\x63om.keepersecurity.protoB\x05Vaultb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'SyncDown_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  _globals['DESCRIPTOR']._options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\030com.keepersecurity.protoB\005Vault'
  _globals['_CACHESTATUS']._serialized_start=6870
  _globals['_CACHESTATUS']._serialized_end=6904
  _globals['_RECORDROTATIONSTATUS']._serialized_start=6906
  _globals['_RECORDROTATIONSTATUS']._serialized_end=7008
  _globals['_SYNCDOWNREQUEST']._serialized_start=120
  _globals['_SYNCDOWNREQUEST']._serialized_end=185
  _globals['_SYNCDOWNRESPONSE']._serialized_start=188
  _globals['_SYNCDOWNRESPONSE']._serialized_end=2365
  _globals['_USERFOLDER']._serialized_start=2368
  _globals['_USERFOLDER']._serialized_end=2514
  _globals['_SHAREDFOLDER']._serialized_start=2517
  _globals['_SHAREDFOLDER']._serialized_end=2858
  _globals['_USERFOLDERSHAREDFOLDER']._serialized_start=2860
  _globals['_USERFOLDERSHAREDFOLDER']._serialized_end=2946
  _globals['_SHAREDFOLDERFOLDER']._serialized_start=2949
  _globals['_SHAREDFOLDERFOLDER']._serialized_end=3136
  _globals['_SHAREDFOLDERKEY']._serialized_start=3138
  _globals['_SHAREDFOLDERKEY']._serialized_end=3246
  _globals['_TEAM']._serialized_start=3249
  _globals['_TEAM']._serialized_end=3572
  _globals['_RECORD']._serialized_start=3575
  _globals['_RECORD']._serialized_end=3766
  _globals['_RECORDLINK']._serialized_start=3768
  _globals['_RECORDLINK']._serialized_end=3866
  _globals['_USERFOLDERRECORD']._serialized_start=3868
  _globals['_USERFOLDERRECORD']._serialized_end=3942
  _globals['_SHAREDFOLDERFOLDERRECORD']._serialized_start=3944
  _globals['_SHAREDFOLDERFOLDERRECORD']._serialized_end=4051
  _globals['_NONSHAREDDATA']._serialized_start=4053
  _globals['_NONSHAREDDATA']._serialized_end=4101
  _globals['_RECORDMETADATA']._serialized_start=4104
  _globals['_RECORDMETADATA']._serialized_end=4391
  _globals['_SHARINGCHANGE']._serialized_start=4393
  _globals['_SHARINGCHANGE']._serialized_end=4443
  _globals['_PROFILE']._serialized_start=4445
  _globals['_PROFILE']._serialized_end=4507
  _globals['_PROFILEPIC']._serialized_start=4509
  _globals['_PROFILEPIC']._serialized_end=4552
  _globals['_PENDINGTEAMMEMBER']._serialized_start=4554
  _globals['_PENDINGTEAMMEMBER']._serialized_end=4666
  _globals['_BREACHWATCHRECORD']._serialized_start=4669
  _globals['_BREACHWATCHRECORD']._serialized_end=4835
  _globals['_USERAUTH']._serialized_start=4838
  _globals['_USERAUTH']._serialized_end=5018
  _globals['_BREACHWATCHSECURITYDATA']._serialized_start=5020
  _globals['_BREACHWATCHSECURITYDATA']._serialized_end=5082
  _globals['_REUSEDPASSWORDS']._serialized_start=5084
  _globals['_REUSEDPASSWORDS']._serialized_end=5134
  _globals['_SHAREDFOLDERRECORD']._serialized_start=5137
  _globals['_SHAREDFOLDERRECORD']._serialized_end=5434
  _globals['_SHAREDFOLDERUSER']._serialized_start=5437
  _globals['_SHAREDFOLDERUSER']._serialized_end=5678
  _globals['_SHAREDFOLDERTEAM']._serialized_start=5681
  _globals['_SHAREDFOLDERTEAM']._serialized_end=5915
  _globals['_KSMCHANGE']._serialized_start=5918
  _globals['_KSMCHANGE']._serialized_end=6056
  _globals['_SHAREINVITATION']._serialized_start=6058
  _globals['_SHAREINVITATION']._serialized_end=6093
  _globals['_USER']._serialized_start=6095
  _globals['_USER']._serialized_end=6139
  _globals['_SYNCDIAGNOSTICS']._serialized_start=6141
  _globals['_SYNCDIAGNOSTICS']._serialized_end=6264
  _globals['_RECORDROTATION']._serialized_start=6267
  _globals['_RECORDROTATION']._serialized_end=6505
  _globals['_SECURITYSCOREDATA']._serialized_start=6507
  _globals['_SECURITYSCOREDATA']._serialized_end=6577
  _globals['_BREACHWATCHGETSYNCDATAREQUEST']._serialized_start=6579
  _globals['_BREACHWATCHGETSYNCDATAREQUEST']._serialized_end=6630
  _globals['_BREACHWATCHGETSYNCDATARESPONSE']._serialized_start=6633
  _globals['_BREACHWATCHGETSYNCDATARESPONSE']._serialized_end=6812
  _globals['_GETACCOUNTUIDMAPRESPONSE']._serialized_start=6814
  _globals['_GETACCOUNTUIDMAPRESPONSE']._serialized_end=6868
# @@protoc_insertion_point(module_scope)



#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Commander
# Copyright 2022 Keeper Security Inc.
# Contact: ops@keepersecurity.com
#

import requests
import os
import json
import logging
import ssl
import time

from typing import Union, Dict, Optional

from .params import RestApiContext
from .error import KeeperApiError, Error
from .proto import APIRequest_pb2 as proto
from . import crypto, utils
from cryptography.hazmat.primitives.asymmetric import rsa, ec

CLIENT_VERSION = 'c17.1.9'

SERVER_PUBLIC_KEYS = {
    1: crypto.load_rsa_public_key(utils.base64_url_decode(
        'MIIBCgKCAQEA9Z_CZzxiNUz8-npqI4V10-zW3AL7-M4UQDdd_17759Xzm0MOEfH' +
        'OOsOgZxxNK1DEsbyCTCE05fd3Hz1mn1uGjXvm5HnN2mL_3TOVxyLU6VwH9EDInn' +
        'j4DNMFifs69il3KlviT3llRgPCcjF4xrF8d4SR0_N3eqS1f9CBJPNEKEH-am5Xb' +
        '_FqAlOUoXkILF0UYxA_jNLoWBSq-1W58e4xDI0p0GuP0lN8f97HBtfB7ijbtF-V' +
        'xIXtxRy-4jA49zK-CQrGmWqIm5DzZcBvUtVGZ3UXd6LeMXMJOifvuCneGC2T2uB' +
        '6G2g5yD54-onmKIETyNX0LtpR1MsZmKLgru5ugwIDAQAB')),

    2: crypto.load_rsa_public_key(utils.base64_url_decode(
        'MIIBCgKCAQEAkOpym7xC3sSysw5DAidLoVF7JUgnvXejbieDWmEiD-DQOKxzfQq' +
        'YHoFfeeix__bx3wMW3I8cAc8zwZ1JO8hyB2ON732JE2Zp301GAUMnAK_rBhQWmY' +
        'KP_-uXSKeTJPiuaW9PVG0oRJ4MEdS-t1vIA4eDPhI1EexHaY3P2wHKoV8twcGvd' +
        'WUZB5gxEpMbx5CuvEXptnXEJlxKou3TZu9uwJIo0pgqVLUgRpW1RSRipgutpUsl' +
        'BnQ72Bdbsry0KKVTlcPsudAnnWUtsMJNgmyQbESPm-aVv-GzdVUFvWKpKkAxDpN' +
        'ArPMf0xt8VL2frw2LDe5_n9IMFogUiSYt156_mQIDAQAB')),

    3: crypto.load_rsa_public_key(utils.base64_url_decode(
        'MIIBCgKCAQEAyvxCWbLvtMRmq57oFg3mY4DWfkb1dir7b29E8UcwcKDcCsGTqoI' +
        'hubU2pO46TVUXmFgC4E-Zlxt-9F-YA-MY7i_5GrDvySwAy4nbDhRL6Z0kz-rqUi' +
        'rgm9WWsP9v-X_BwzARqq83HNBuzAjf3UHgYDsKmCCarVAzRplZdT3Q5rnNiYPYS' +
        'HzwfUhKEAyXk71UdtleD-bsMAmwnuYHLhDHiT279An_Ta93c9MTqa_Tq2Eirl_N' +
        'Xn1RdtbNohmMXldAH-C8uIh3Sz8erS4hZFSdUG1WlDsKpyRouNPQ3diorbO88wE' +
        'AgpHjXkOLj63d1fYJBFG0yfu73U80aEZehQkSawIDAQAB')),

    4: crypto.load_rsa_public_key(utils.base64_url_decode(
        'MIIBCgKCAQEA0TVoXLpgluaqw3P011zFPSIzWhUMBqXT-Ocjy8NKjJbdrbs53eR' +
        'FKk1waeB3hNn5JEKNVSNbUIe-MjacB9P34iCfKtdnrdDB8JXx0nIbIPzLtcJC4H' +
        'CYASpjX_TVXrU9BgeCE3NUtnIxjHDy8PCbJyAS_Pv299Q_wpLWnkkjq70ZJ2_fX' +
        '-ObbQaZHwsWKbRZ_5sD6rLfxNACTGI_jo9-vVug6AdNq96J7nUdYV1cG-INQwJJ' +
        'KMcAbKQcLrml8CMPc2mmf0KQ5MbS_KSbLXHUF-81AsZVHfQRSuigOStQKxgSGL5' +
        'osY4NrEcODbEXtkuDrKNMsZYhijKiUHBj9vvgKwIDAQAB')),

    5: crypto.load_rsa_public_key(utils.base64_url_decode(
        'MIIBCgKCAQEAueOWC26w-HlOLW7s88WeWkXpjxK4mkjqngIzwbjnsU9145R51Hv' +
        'sILvjXJNdAuueVDHj3OOtQjfUM6eMMLr-3kaPv68y4FNusvB49uKc5ETI0HtHmH' +
        'FSn9qAZvC7dQHSpYqC2TeCus-xKeUciQ5AmSfwpNtwzM6Oh2TO45zAqSA-QBSk_' +
        'uv9TJu0e1W1AlNmizQtHX6je-mvqZCVHkzGFSQWQ8DBL9dHjviI2mmWfL_egAVV' +
        'hBgTFXRHg5OmJbbPoHj217Yh-kHYA8IWEAHylboH6CVBdrNL4Na0fracQVTm-nO' +
        'WdM95dKk3fH-KJYk_SmwB47ndWACLLi5epLl9vwIDAQAB')),

    6: crypto.load_rsa_public_key(utils.base64_url_decode(
        'MIIBCgKCAQEA2PJRM7-4R97rHwY_zCkFA8B3llawb6gF7oAZCpxprl6KB5z2cqL' +
        'AvUfEOBtnr7RIturX04p3ThnwaFnAR7ADVZWBGOYuAyaLzGHDI5mvs8D-NewG9v' +
        'w8qRkTT7Mb8fuOHC6-_lTp9AF2OA2H4QYiT1vt43KbuD0Y2CCVrOTKzDMXG8msl' +
        '_JvAKt4axY9RGUtBbv0NmpkBCjLZri5AaTMgjLdu8XBXCqoLx7qZL-Bwiv4njw-' +
        'ZAI4jIszJTdGzMtoQ0zL7LBj_TDUBI4Qhf2bZTZlUSL3xeDWOKmd8Frksw3oKyJ' +
        '17oCQK-EGau6EaJRGyasBXl8uOEWmYYgqOWirNwIDAQAB')),

    7: crypto.load_ec_public_key(utils.base64_url_decode(
        'BK9w6TZFxE6nFNbMfIpULCup2a8xc6w2tUTABjxny7yFmxW0dAEojwC6j6zb5nTlmb1dAx8nwo3qF7RPYGmloRM')),

    8: crypto.load_ec_public_key(utils.base64_url_decode(
        'BKnhy0obglZJK-igwthNLdknoSXRrGB-mvFRzyb_L-DKKefWjYdFD2888qN1ROczz4n3keYSfKz9Koj90Z6w_tQ')),

    9: crypto.load_ec_public_key(utils.base64_url_decode(
        'BAsPQdCpLIGXdWNLdAwx-3J5lNqUtKbaOMV56hUj8VzxE2USLHuHHuKDeno0ymJt-acxWV1xPlBfNUShhRTR77g')),

    10: crypto.load_ec_public_key(utils.base64_url_decode(
        'BNYIh_Sv03nRZUUJveE8d2mxKLIDXv654UbshaItHrCJhd6cT7pdZ_XwbdyxAOCWMkBb9AZ4t1XRCsM8-wkEBRg')),

    11: crypto.load_ec_public_key(utils.base64_url_decode(
        'BA6uNfeYSvqagwu4TOY6wFK4JyU5C200vJna0lH4PJ-SzGVXej8l9dElyQ58_ljfPs5Rq6zVVXpdDe8A7Y3WRhk')),

    12: crypto.load_ec_public_key(utils.base64_url_decode(
        'BMjTIlXfohI8TDymsHxo0DqYysCy7yZGJ80WhgOBR4QUd6LBDA6-_318a-jCGW96zxXKMm8clDTKpE8w75KG-FY')),

    13: crypto.load_ec_public_key(utils.base64_url_decode(
        'BJBDU1P1H21IwIdT2brKkPqbQR0Zl0TIHf7Bz_OO9jaNgIwydMkxt4GpBmkYoprZ_DHUGOrno2faB7pmTR7HhuI')),

    14: crypto.load_ec_public_key(utils.base64_url_decode(
        'BJFF8j-dH7pDEw_U347w2CBM6xYM8Dk5fPPAktjib-opOqzvvbsER-WDHM4ONCSBf9O_obAHzCyygxmtpktDuiE')),

    15: crypto.load_ec_public_key(utils.base64_url_decode(
        'BDKyWBvLbyZ-jMueORl3JwJnnEpCiZdN7yUvT0vOyjwpPBCDf6zfL4RWzvSkhAAFnwOni_1tQSl8dfXHbXqXsQ8')),

    16: crypto.load_ec_public_key(utils.base64_url_decode(
        'BDXyZZnrl0tc2jdC5I61JjwkjK2kr7uet9tZjt8StTiJTAQQmnVOYBgbtP08PWDbecxnHghx3kJ8QXq1XE68y8c')),

    17: crypto.load_ec_public_key(utils.base64_url_decode(
        'BFX68cb97m9_sweGdOVavFM3j5ot6gveg6xT4BtGahfGhKib-zdZyO9pwvv1cBda9ahkSzo1BQ4NVXp9qRyqVGU')),
}   # type: Dict[int, Union[rsa.RSAPublicKey, ec.EllipticCurvePublicKey]]


def encrypt_with_keeper_key(context, data: bytes) -> bytes:
    key_id = context.server_key_id
    if 1 <= key_id <= 6:
        return crypto.encrypt_rsa(data, SERVER_PUBLIC_KEYS[key_id])
    elif 7 <= key_id <= 17:
        return crypto.encrypt_ec(data, SERVER_PUBLIC_KEYS[key_id])
    else:
        raise KeeperApiError('invalid_key_id', f'Key ID \"{key_id}\" is not valid.')


def execute_rest(context, endpoint, payload):
    # type: (RestApiContext, str, proto.ApiRequestPayload) -> Optional[Union[bytes, dict]]
    if not context.transmission_key:
        context.transmission_key = os.urandom(32)

    if not context.server_key_id:
        context.server_key_id = 7

    run_request = True
    while run_request:
        run_request = False

        api_request = proto.ApiRequest()
        server_public_key = SERVER_PUBLIC_KEYS[context.server_key_id]
        if isinstance(server_public_key, rsa.RSAPublicKey):
            api_request.encryptedTransmissionKey = crypto.encrypt_rsa(context.transmission_key, server_public_key)
        elif isinstance(server_public_key, ec.EllipticCurvePublicKey):
            api_request.encryptedTransmissionKey = crypto.encrypt_ec(context.transmission_key, server_public_key)
        else:
            raise ValueError('Invalid server public key')
        api_request.publicKeyId = context.server_key_id
        api_request.locale = context.locale or 'en_US'

        api_request.encryptedPayload = crypto.encrypt_aes_v2(payload.SerializeToString(), context.transmission_key)

        request_data = api_request.SerializeToString()
        if endpoint.startswith('https://'):
            url = endpoint
        else:
            url = context.server_base + endpoint

        try:
            rs = requests.post(url, data=request_data, headers={'Content-Type': 'application/octet-stream'},
                               proxies=context.proxies, verify=context.certificate_check)
        except requests.exceptions.SSLError as e:
            doc_url = 'https://docs.keeper.io/secrets-manager/commander-cli/using-commander/troubleshooting-commander-cli#ssl-certificate-errors'
            if len(e.args) > 0:
                inner_e = e.args[0]
                if hasattr(inner_e, 'reason'):
                    reason = getattr(inner_e, 'reason')
                    if isinstance(reason, Exception) and hasattr(reason, 'args'):
                        args = getattr(reason, 'args')
                        if isinstance(args, tuple) and len(args) > 0:
                            inner_e = args[0]
                            if isinstance(inner_e, ssl.SSLCertVerificationError):
                                raise Error(f'Certificate validation error. More info:\n{doc_url}')
            raise e

        content_type = rs.headers.get('Content-Type') or ''
        if rs.status_code == 200:
            if content_type == 'application/json':
                return rs.json()

            rs_body = rs.content
            if rs_body:
                rs_body = crypto.decrypt_aes_v2(rs.content, context.transmission_key)
            return rs_body
        elif rs.status_code >= 400:
            if content_type.startswith('application/json'):
                failure = rs.json()
                logging.debug('<<< Response Error: [%s]', failure)
                if rs.status_code == 401:
                    if failure.get('error') == 'key':
                        server_key_id = failure['key_id']
                        if server_key_id != context.server_key_id:
                            context.server_key_id = server_key_id
                            run_request = True
                            continue
                elif rs.status_code == 403:
                    if failure.get('error') == 'throttled' and not context.fail_on_throttle:
                        logging.info('Throttled. sleeping for 10 seconds')
                        time.sleep(10)
                        run_request = True
                        continue
                return failure
            else:
                if logging.getLogger().level <= logging.DEBUG:
                    if rs.text:
                        logging.debug('<<< Response Content: [%s]', rs.text)
                    else:
                        logging.debug('<<< HTTP Status: [%s]  Reason: [%s]', rs.status_code, rs.reason)
                raise KeeperApiError(rs.status_code, rs.reason)


def v2_execute(context, rq):
    # type: (RestApiContext, dict) -> Optional[dict]

    api_request_payload = proto.ApiRequestPayload()
    api_request_payload.payload = json.dumps(rq).encode('utf-8')
    rs_data = execute_rest(context, 'vault/execute_v2_command', api_request_payload)
    if rs_data:
        if type(rs_data) is bytes:
            rs = json.loads(rs_data.decode('utf-8'))
            logger = logging.getLogger()
            if logger.level <= logging.DEBUG:
                logger.debug('>>> Request JSON: [%s]', json.dumps(rq, sort_keys=True, indent=4))
                logger.debug('<<< Response JSON: [%s]', json.dumps(rs, sort_keys=True, indent=4))
            return rs

        if type(rs_data) is dict:
            raise KeeperApiError(rs_data['error'], rs_data['message'])



import sys
import unittest
from unittest import mock

from keepercommander.error import CommandError

if sys.version_info >= (3, 8):
    import datetime
    import socket
    import string
    from cryptography import x509
    from cryptography.hazmat._oid import NameOID
    from cryptography.hazmat.backends import default_backend
    from cryptography.hazmat.primitives import serialization, hashes
    from cryptography.hazmat.primitives.asymmetric import ec

    from keepercommander.commands.tunnel.port_forward.tunnel_helpers import (generate_random_bytes, find_open_port)

    def generate_self_signed_cert(private_key):
        # Generate a self-signed certificate
        subject = issuer = x509.Name([
            x509.NameAttribute(NameOID.COMMON_NAME, u"localhost"),
        ])
        cert = (
            x509.CertificateBuilder()
            .subject_name(subject)
            .issuer_name(issuer)
            .public_key(private_key.public_key())
            .serial_number(x509.random_serial_number())
            .not_valid_before(datetime.datetime.utcnow())
            .not_valid_after(
                # Our certificate will be valid for 10 days
                datetime.datetime.utcnow() + datetime.timedelta(days=10)
            )
            .sign(private_key, hashes.SHA256(), default_backend())
        )
        cert_pem = cert.public_bytes(serialization.Encoding.PEM).decode('utf-8')

        return cert_pem


    def new_private_key():
        # Generate an EC private key
        private_key = ec.generate_private_key(
            ec.SECP256R1(),  # Using P-256 curve
            backend=default_backend()
        )
        # Serialize to PEM format
        private_key_str = private_key.private_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PrivateFormat.PKCS8,
            encryption_algorithm=serialization.NoEncryption()
        ).decode('utf-8')
        return private_key, private_key_str


    class TestFindOpenPort(unittest.TestCase):
        def mock_bind(self, address):
            # Mock the behavior of socket.socket.bind
            port = address[1]
            if port in self.in_use_ports:
                raise OSError("Address already in use")
            else:
                print(f"Port {port} bound successfully.")

        def test_preferred_port(self):
            # Test that the function returns the preferred port if it's available
            preferred_port = 50000
            open_port = find_open_port([], preferred_port=preferred_port)
            self.assertEqual(open_port, preferred_port)

        def test_preferred_port_unavailable(self):
            # Mock the bind method to simulate that port 80 is in use
            with mock.patch('socket.socket.bind', side_effect=OSError("Address already in use")):
                preferred_port = 80
                with self.assertRaises(CommandError):
                    open_port = find_open_port([], preferred_port=preferred_port)

        def test_range(self):
            # Test that the function returns a port within the specified range
            start_port = 50000
            end_port = 50010
            open_port = find_open_port([], start_port=start_port, end_port=end_port)
            self.assertTrue(start_port <= open_port <= end_port)

        def test_no_available_ports(self):
            # Setup
            self.in_use_ports = set(range(50000, 50011))  # All these ports are in use

            # Patch
            with mock.patch.object(socket.socket, 'bind', side_effect=self.mock_bind):
                # Test
                open_port = find_open_port([], start_port=50000, end_port=50010)
                self.assertIsNone(open_port)

        def test_invalid_range(self):
            # Test that the function returns None if the range is invalid
            open_port = find_open_port([], start_port=50010, end_port=50000)
            self.assertIsNone(open_port)

        def test_socket_exception(self):
            # Test that the function handles exceptions other than OSError gracefully
            with mock.patch('socket.socket.bind', side_effect=Exception("Test exception")):
                open_port = find_open_port([], start_port=49152, end_port=49153, host='localhost')
                self.assertIsNone(open_port)

        def test_tried_ports(self):
            # Setup
            self.in_use_ports = {50000, 50001}  # These ports are in use

            # Patch
            with mock.patch.object(socket.socket, 'bind', side_effect=self.mock_bind):
                # Test
                open_port = find_open_port([50000, 50001], start_port=50000, end_port=50002)
                self.assertEqual(open_port, 50002)


    class TestGenerateRandomBytes(unittest.TestCase):

        def test_default_length(self):
            # Test that the default length of the returned bytes is 32
            random_bytes = generate_random_bytes()
            self.assertEqual(len(random_bytes), 32, f'Length 32 failed found {len(random_bytes)} in '
                                                    f'{random_bytes}')

        def test_custom_length(self):
            # Test custom lengths
            for length in [1, 10, 20, 50, 100]:
                random_bytes = generate_random_bytes(length)
                self.assertEqual(len(random_bytes), length, f'Length {length} failed found {len(random_bytes)} in '
                                                            f'{random_bytes}')

        def test_content(self):
            # Test that the returned bytes only contain printable characters
            for length in [1, 10, 20, 50, 100]:
                random_bytes = generate_random_bytes(length)
                self.assertTrue(all(byte in string.printable.encode('utf-8') for byte in random_bytes))

        def test_zero_length(self):
            # Test that a zero length returns an empty bytes object
            random_bytes = generate_random_bytes(0)
            self.assertEqual(random_bytes, b'')

        def test_negative_length(self):
            # Test that a negative length raises a ValueError
            with self.assertRaises(ValueError):
                generate_random_bytes(-1)

        def test_type(self):
            # Test that the return type is bytes
            random_bytes = generate_random_bytes()
            self.assertIsInstance(random_bytes, bytes)

        def test_uniqueness(self):
            # Test that multiple calls return different values
            random_bytes1 = generate_random_bytes()
            random_bytes2 = generate_random_bytes()
            self.assertNotEqual(random_bytes1, random_bytes2)



# -*- coding: utf-8 -*-
# flake8: noqa
# noreorder
"""
Underscore-to-camelCase converter (and vice versa) for strings and dict keys in Python.
"""
__title__ = "pyhumps"
__version__ = "1.6.1"
__author__ = "Nick Ficano"
__license__ = "MIT License"
__copyright__ = "Copyright 2019 Nick Ficano"

from .humps import decamelize, depascalize

__all__ = ['decamelize', 'depascalize']


#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Commander
# Copyright 2025 Keeper Security Inc.
# Contact: ops@keepersecurity.com
#

import logging
import os
import platform
import subprocess
from typing import Dict, Any, Tuple, Optional

from fido2.webauthn import PublicKeyCredentialCreationOptions, PublicKeyCredentialRequestOptions

from .... import utils
from ..base import BasePlatformHandler, StorageHandler
from .keychain import MacOSKeychainManager
from .webauthn import MacOSTouchIDWebAuthnClient
from ...utils.constants import (
    MACOS_PREFS_PATH,
    ERROR_MESSAGES,
    PLATFORM_DARWIN,
    DEFAULT_BIOMETRIC_TIMEOUT
)
from ...utils.error_handler import BiometricErrorHandler

MACOS_BIOUTIL_COMMAND = ['bioutil', '-r', '-s']

class MacOSStorageHandler(StorageHandler):
    """macOS plist storage handler"""

    def __init__(self):
        self.prefs_path = self._get_prefs_path()

    def _get_prefs_path(self):
        """Get macOS preferences path"""
        home_dir = os.path.expanduser("~")
        return os.path.join(home_dir, "Library", "Preferences", MACOS_PREFS_PATH)

    def get_biometric_flag(self, username: str) -> bool:
        """Get biometric flag from macOS preferences - True if credential ID exists"""
        return self.get_credential_id(username) is not None

    def delete_biometric_flag(self, username: str) -> bool:
        """Delete biometric flag from macOS preferences - removes credential ID"""
        return self.delete_credential_id(username)

    def store_credential_id(self, username: str, credential_id: str) -> bool:
        """Store credential ID for user in macOS preferences (also serves as biometric flag)"""
        try:
            import plistlib
            prefs = {}
            
            if os.path.exists(self.prefs_path):
                try:
                    with open(self.prefs_path, 'rb') as f:
                        prefs = plistlib.load(f)
                except Exception:
                    prefs = {}
            
            prefs[username] = credential_id
            
            os.makedirs(os.path.dirname(self.prefs_path), exist_ok=True)
            with open(self.prefs_path, 'wb') as f:
                plistlib.dump(prefs, f)
            
            logging.debug("Stored credential ID for user: %s", username)
            return True
        except Exception as e:
            logging.warning("Failed to store credential ID for %s: %s", username, str(e))
            BiometricErrorHandler.create_storage_error("store credential ID", "macOS", e)
            return False

    def get_credential_id(self, username: str) -> Optional[str]:
        """Get stored credential ID for user from macOS preferences"""
        try:
            import plistlib
            if not os.path.exists(self.prefs_path):
                return None
            
            with open(self.prefs_path, 'rb') as f:
                prefs = plistlib.load(f)
            
            value = prefs.get(username)
            
            if value:
                if isinstance(value, str):
                    return value
            
            return None
        except Exception as e:
            logging.warning("Failed to retrieve credential ID for %s: %s", username, str(e))
            BiometricErrorHandler.create_storage_error("get credential ID", "macOS", e)
            return None

    def delete_credential_id(self, username: str) -> bool:
        """Delete stored credential ID for user from macOS preferences"""
        try:
            import plistlib
            if not os.path.exists(self.prefs_path):
                return True 
            
            with open(self.prefs_path, 'rb') as f:
                prefs = plistlib.load(f)
            
            if username not in prefs:
                return True
                
            del prefs[username]
            
            with open(self.prefs_path, 'wb') as f:
                plistlib.dump(prefs, f)
            
            with open(self.prefs_path, 'rb') as f:
                verification_prefs = plistlib.load(f)
            
            return username not in verification_prefs
        except Exception as e:
            logging.warning("Failed to delete credential ID for %s: %s", username, str(e))
            BiometricErrorHandler.create_storage_error("delete credential ID", "macOS", e)
            return False

class MacOSHandler(BasePlatformHandler):
    """macOS-specific biometric handler"""

    def __init__(self):
        super().__init__()
        self.keychain_manager = MacOSKeychainManager()

    def _create_storage_handler(self) -> StorageHandler:
        return MacOSStorageHandler()

    def _get_platform_name(self) -> str:
        return "Touch ID"

    def detect_capabilities(self) -> Tuple[bool, str]:
        """Detect Touch ID availability on macOS"""
        if platform.system() != PLATFORM_DARWIN:
            return False, "Not running on macOS"

        error_messages = []

        try:
            # Try bioutil command first
            if self._check_bioutil_command(error_messages):
                return True, "Touch ID is available and configured"

            # Fallback: LocalAuthentication check
            if self._check_local_authentication(error_messages):
                return True, "Touch ID is available"

            # If we get here, all detection methods failed
            detailed_error = "Touch ID detection failed. " + "; ".join(error_messages)
            detailed_error += ". Please verify Touch ID is set up in System Preferences > Touch ID & Password"
            return False, detailed_error

        except Exception as e:
            return False, f"Error checking Touch ID: {str(e)}"

    def _check_bioutil_command(self, error_messages: list) -> bool:
        """Check Touch ID using bioutil command"""
        try:
            result = subprocess.run(MACOS_BIOUTIL_COMMAND, capture_output=True, text=True, timeout=10)

            if result.returncode == 0:
                output = result.stdout.lower()
                if ('touch id' in output or 
                    'biometrics functionality: 1' in output or
                    'biometric' in output):
                    return True
                else:
                    error_messages.append("bioutil: ran successfully but no Touch ID detected")
            else:
                error_messages.append(f"bioutil: command failed (return code {result.returncode})")
        except FileNotFoundError:
            error_messages.append("bioutil: command not found")
        except Exception as e:
            error_messages.append(f"bioutil: {str(e)}")

        return False

    def _check_local_authentication(self, error_messages: list) -> bool:
        """Check Touch ID using LocalAuthentication framework"""
        try:
            import LocalAuthentication  # pylint: disable=import-error
            context = LocalAuthentication.LAContext.alloc().init()  # pylint: disable=no-member
            error = None
            
            policy_attr = getattr(LocalAuthentication, 'LAPolicyDeviceOwnerAuthenticationWithBiometrics', None)
            if policy_attr is None:
                error_messages.append("LocalAuthentication: biometric policy not available")
                return False
                
            can_evaluate = context.canEvaluatePolicy_error_(policy_attr, error)
            
            if can_evaluate:
                return True
            else:
                la_error = "LocalAuthentication: policy evaluation failed"
                if error:
                    la_error += f" (error: {error})"
                error_messages.append(la_error)
                
        except ImportError as e:
            error_messages.append(f"LocalAuthentication: import failed - {str(e)}")
            error_messages.append("LocalAuthentication: try 'pip install pyobjc-framework-LocalAuthentication'")
        except Exception as e:
            error_messages.append(f"LocalAuthentication: {str(e)}")

        return False

    def create_webauthn_client(self, data_collector):
        """Create macOS Touch ID WebAuthn client"""
        try:
            return MacOSTouchIDWebAuthnClient(data_collector, self.keychain_manager)
        except ImportError:
            raise Exception('macOS Touch ID client dependencies not available')

    def handle_credential_creation(self, creation_options: Dict[str, Any]) -> Dict[str, Any]:
        """Handle macOS-specific credential creation"""
        rp_id = creation_options.get('rp', {}).get('id')
        if not rp_id:
            raise Exception("No RP ID found in creation options - server configuration error")
            
        # Check for existing credentials before processing
        if 'excludeCredentials' in creation_options and creation_options['excludeCredentials']:
            for excluded_cred in creation_options['excludeCredentials']:
                cred_id = excluded_cred.get('id')
                if isinstance(cred_id, str):
                    cred_id_b64 = cred_id
                else:
                    cred_id_b64 = utils.base64_url_encode(cred_id)
                
                if self.keychain_manager.credential_exists(cred_id_b64, rp_id, DEFAULT_BIOMETRIC_TIMEOUT):
                    raise Exception(ERROR_MESSAGES['credential_exists'])

        return self._prepare_credential_creation_options(creation_options)

    def handle_authentication_options(self, pk_options: Dict[str, Any]) -> Dict[str, Any]:
        """Handle macOS-specific authentication options"""
        return self._prepare_authentication_options(pk_options)

    def perform_authentication(self, client, options: PublicKeyCredentialRequestOptions):
        """Perform macOS Touch ID authentication"""
        try:
            return client.get_assertion(options)
        except Exception as e:
            raise self._handle_authentication_error(e, self._get_platform_name())

    def perform_credential_creation(self, client, options: PublicKeyCredentialCreationOptions):
        """Perform macOS Touch ID credential creation"""
        try:
            return client.make_credential(options)
        except Exception as e:
            raise self._handle_credential_creation_error(e, self._get_platform_name()) 


#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Commander
# Copyright 2024 Keeper Security Inc.
# Contact: ops@keepersecurity.com
#

from ...service.app import create_app
from ...service.config.service_config import ServiceConfig
from ...service.core.service_manager import ServiceManager

flask_app = create_app()


if __name__ == '__main__':
    service_config = ServiceConfig()
    config_data = service_config.load_config()
    
    try:
        from ...service.core.globals import ensure_params_loaded
        print("Pre-loading Keeper parameters for background mode...")
        ensure_params_loaded()
        print("Keeper parameters loaded successfully")
    except Exception as e:
        print(f"Warning: Failed to pre-load parameters during startup: {e}")
        print("Parameters will be loaded on first API call if needed")
    
    ssl_context = None
    
    if not (port := config_data.get("port")):
        print("Error: Service configuration is incomplete. Please configure the service port in service_config")

    ssl_context = ServiceManager.get_ssl_context(config_data)
    
    flask_app.run(
        host='0.0.0.0',
        port=port,
        ssl_context=ssl_context
    )




from enum import Enum
from typing import Dict, List

from . import api, utils
from .proto import enterprise_pb2
from .record import Record
from .subfolder import find_folders, get_folder_path


def get_shared_records(params, record_uids, cache_only=False):
    def fetch_team_members(t_uids):
        members = {}
        if params.enterprise_ec_key:
            for team_uid in t_uids:
                team_users = members.get(team_uid, set())
                rq = enterprise_pb2.GetTeamMemberRequest()
                rq.teamUid = utils.base64_url_decode(team_uid)
                endpoint = 'vault/get_team_members'
                rs = api.communicate_rest(params, rq, endpoint, rs_type=enterprise_pb2.GetTeamMemberResponse)
                if rs.enterpriseUser:
                    team_users.update({x.email for x in rs.enterpriseUser})
                    members[team_uid] = team_users
        return members

    def get_cached_team_members(t_uids, uname_lookup):
        members = {}
        if not params.enterprise:
            return members

        team_users = params.enterprise.get('team_users') or []
        team_users = [tu for tu in team_users if tu.get('user_type') != 2 and tu.get('team_uid') in t_uids]

        for tu in team_users:
            user_id = tu.get('enterprise_user_id')
            username = uname_lookup.get(user_id)
            team_uid = tu.get('team_uid')
            t_members = members.get(team_uid, set())
            t_members.add(username)
            members[team_uid] = t_members

        return members

    def fetch_sf_admins():
        sf_uids = [uid for uid in params.shared_folder_cache]
        return {sf_uid: api.get_share_admins_for_shared_folder(params, sf_uid) for sf_uid in sf_uids}

    def get_restricted_role_members(uname_lookup):
        # Get team_uids and usernames (assigned directly and indirectly) in share-restricted roles
        members = set()
        if not params.enterprise:
            return members

        restrict_key = 'restrict_sharing_all'
        enf_key = 'enforcements'
        r_enforcements = params.enterprise.get('role_enforcements', [])
        no_share_roles = {re.get('role_id') for re in r_enforcements if re.get(enf_key).get(restrict_key) == 'true'}
        r_users = [u for u in params.enterprise.get('role_users', []) if u.get('role_id') in no_share_roles]
        r_teams = [t for t in params.enterprise.get('role_teams', []) if t.get('role_id') in no_share_roles]
        no_share_users = {uname_lookup.get(u.get('enterprise_user_id')) for u in r_users}
        no_share_teams = {t.get('team_uid') for t in r_teams}
        cached_team_members = get_cached_team_members(no_share_teams, uname_lookup)
        no_share_team_members = {t for team_uid in no_share_teams for t in cached_team_members.get(team_uid) or []}
        members = no_share_users.union(no_share_teams).union(no_share_team_members)
        return members

    api.get_record_shares(params, record_uids)
    sf_teams = [shared_folder.get('teams', []) for shared_folder in params.shared_folder_cache.values()]
    sf_share_admins = fetch_sf_admins() if not cache_only else {}
    team_uids = {t.get('team_uid') for teams in sf_teams for t in teams}
    enterprise_users = params.enterprise.get('users') if params.enterprise else []
    username_lookup = {u.get('enterprise_user_id'): u.get('username') for u in enterprise_users}
    restricted_role_members = get_restricted_role_members(username_lookup)
    team_members = get_cached_team_members(team_uids, username_lookup) if cache_only or params.enterprise \
        else fetch_team_members(team_uids)
    records = [api.get_record(params, uid) for uid in record_uids]  # type: List[Record or None]
    records = [r for r in records if r]
    shared_records = [SharedRecord(params, r, sf_share_admins, team_members, restricted_role_members) for r in records]
    return {shared_record.uid: shared_record for shared_record in shared_records}


class SharePermissions:
    SharePermissionsType = Enum('SharePermissionsType', ['USER', 'SF_USER', 'TEAM', 'TEAM_USER'])
    bits_text_lookup = {(1 << 0): 'Edit', (1 << 1): 'Share'}

    def __init__(self, sp_types=None):
        self.to_uid = ''
        self.to_name = ''
        self.can_edit = False
        self.can_share = False
        self.can_view = True
        self.expiration = 0
        self.folder_path = ''
        self.types = set()
        self.bits = 0
        self.is_admin = False
        self.team_members = dict()
        self.user_perms: Dict[str, SharePermissions] = {}
        self.team_perms: Dict[str, SharePermissions] = {}
        self.update_types(sp_types)

    def update_types(self, sp_types):
        if sp_types is not None:
            update_types_fn = self.types.update if isinstance(sp_types, set) else self.types.add
            update_types_fn(sp_types)

    def get_target(self, show_team_info):
        return self.get_team_view_name() if show_team_info else self.to_name

    def get_team_view_name(self):
        prefix_lookup = {
            SharePermissions.SharePermissionsType.TEAM: '(Team)',
            SharePermissions.SharePermissionsType.TEAM_USER: '(Team User)',
            SharePermissions.SharePermissionsType.USER: '',
            SharePermissions.SharePermissionsType.SF_USER: ''
        }
        prefix = ''.join(prefix_lookup.get(t) for t in self.types)
        return f'{prefix} {self.to_name}'.strip()

    @property
    def permissions_text(self):
        if not self.can_edit and not self.can_share:
            return 'Read Only' if self.can_view else 'Launch Only'
        else:
            privs = [self.can_share and 'Share', self.can_edit and 'Edit']
            return f'Can {" & ".join([p for p in privs if p])}'

    @staticmethod
    def load_permissions(perms, sp_type):
        sp = SharePermissions(sp_type)
        sp.to_uid = perms.get('uid') or perms.get('team_uid')
        sp.to_name = perms.get('username') or perms.get('name')
        sp.is_admin = perms.get('share_admin') or perms.get('is_admin')
        sp.can_edit = perms.get('editable') or perms.get('manage_records') or sp.is_admin
        sp.can_share = perms.get('shareable') or perms.get('manage_users') or sp.is_admin
        sp.can_view = perms.get('view', True)
        exp = perms.get('expiration')
        if isinstance(exp, int) and exp > 0:
            sp.expiration = exp
        return sp

    def apply_restrictions(self, *restrictions):
        for member in self.team_members.values():
            member.apply_restrictions(*restrictions)
        restrictions = ','.join(restrictions).lower()
        if 'edit' in restrictions:
            self.can_edit = False
        if 'share' in restrictions:
            self.can_share = False
        if 'view' in restrictions:
            self.can_view = False


class SharedRecord:
    """Defines a Keeper Shared Record (shared either via Direct-Share or as a child of a Shared-Folder node)"""

    def __init__(self, params, record, sf_sharing_admins, team_members, role_restricted_members):
        has_owner = record.record_uid in params.record_owner_cache
        user_owned = has_owner and params.record_owner_cache.get(record.record_uid).owner
        self.owner = params.user if user_owned else ''
        self.uid = record.record_uid
        self.name = record.title
        self.shared_folders = None
        self.sf_shares = {}
        self.permissions: Dict[str, SharePermissions] = {}
        self.team_permissions: Dict[str, SharePermissions] = {}
        self.user_permissions: Dict[str, SharePermissions] = {}
        self.revision = None
        self.params = params
        self.folder_uids = list(find_folders(params, record.record_uid))
        self.folder_paths = [get_folder_path(params, fuid) for fuid in self.folder_uids]
        self.team_members = team_members

        self.load(params, sf_sharing_admins, team_members, role_restricted_members)

    def get_ordered_permissions(self):
        ordered = list(self.permissions.values())
        for user_perms in self.user_permissions.values():
            if user_perms.to_uid:
                ordered.remove(user_perms)
                team_perms = self.team_permissions.get(user_perms.to_uid)
                ordered.insert(ordered.index(team_perms) + 1, user_perms)
        return ordered

    def merge_permissions(self, share_target, perms_to_merge, sp_type):
        new_perms = SharePermissions.load_permissions(perms_to_merge, sp_type)
        existing = self.permissions.get(share_target) or new_perms
        existing.to_uid = new_perms.to_uid or existing.to_uid
        existing.is_admin = existing.is_admin or new_perms.is_admin
        existing.can_share = existing.can_share or new_perms.can_share
        existing.can_edit = existing.can_edit or new_perms.can_edit
        existing.update_types(new_perms.types)
        if existing.expiration > 0:
            if new_perms.expiration > 0:
                if new_perms.expiration > existing.expiration:
                    existing.expiration = new_perms.expiration
            else:
                existing.expiration = 0

        self.permissions[share_target] = existing
        return existing

    def merge_user_permissions(self, email, user_perms, sp_type=None):
        new_perms = self.merge_permissions(email, user_perms, sp_type or SharePermissions.SharePermissionsType.USER)
        self.user_permissions[email] = new_perms
        return new_perms

    def merge_team_permissions(self, team_uid, team_perms):
        new_perms = self.merge_permissions(team_uid, team_perms, SharePermissions.SharePermissionsType.TEAM)
        self.team_permissions[team_uid] = new_perms
        return new_perms

    def load(self, params, sf_sharing_admins, team_members, role_restricted_members):
        def apply_team_restrictions(team_perms):
            if not params.enterprise:
                return team_perms

            restriction_permission_lookup = {
                'restrict_edit': 'manage_records',
                'restrict_sharing': 'manage_users',
                'restrict_view': 'view'
            }

            teams_cache = params.enterprise.get('teams', {})
            perms = team_perms.copy()
            team_info = next((t for t in teams_cache if t.get('team_uid') == perms.get('team_uid')))
            for restriction, permission in restriction_permission_lookup.items():
                if team_info.get(restriction):
                    perms[permission] = False
            return perms

        def apply_role_restrictions():
            for restricted_target in role_restricted_members:
                perms = self.permissions.get(restricted_target)
                perms and perms.apply_restrictions('share')

        def update_sf_shares(share_to, sf_uid):
            if sf_uid:
                sf_shares = self.sf_shares.get(sf_uid, set())
                sf_shares.add(share_to)
                self.sf_shares[sf_uid] = sf_shares

        def load_user_permissions(u_perms, sf_uid=None, sp_type=None):
            for up in u_perms:
                email = up.get('username')
                update_sf_shares(email, sf_uid)
                share_admins = sf_sharing_admins.get(sf_uid)
                is_admin = share_admins and email in share_admins
                if is_admin:
                    up['editable'] = True
                    up['shareable'] = True
                    up['share_admin'] = True
                self.merge_user_permissions(email, up, sp_type)

        def load_team_permissions(t_perms, sf_uid):
            for tp in t_perms:
                team_uid = tp.get('team_uid')
                team_name = tp.get('name')
                update_sf_shares(team_name, sf_uid)
                tp = apply_team_restrictions(tp)
                merged = self.merge_team_permissions(team_uid, tp)

                # load team-members' permissions
                t_users = team_members.get(team_uid, set())
                ups = [{**tp, 'username': t_username} for t_username in t_users]
                load_user_permissions(ups, sf_uid, SharePermissions.SharePermissionsType.TEAM_USER)
                merged.team_members.update({t_username: self.permissions.get(t_username) for t_username in t_users})

        rec_cached = params.record_cache.get(self.uid)
        shares = rec_cached.get('shares', dict())

        user_perms = list(shares.get('user_permissions', []))
        if len(user_perms) > 0:
            self.owner = next((up.get('username') for up in user_perms if up.get('owner')), '')
            load_user_permissions(user_perms)

        sf_perms = shares.get('shared_folder_permissions', [])
        SF_UID = 'shared_folder_uid'
        sf_cache = params.shared_folder_cache
        shared_folders = {sfp.get(SF_UID): sf_cache.get(sfp.get(SF_UID)) for sfp in sf_perms}
        shared_folders = {k: v for k, v in shared_folders.items() if v}

        sf_user_perms = {sf_uid: sf.get('users') for sf_uid, sf in shared_folders.items() if sf.get('users')}
        team_perms = {sf_uid: sf.get('teams') for sf_uid, sf in shared_folders.items() if sf.get('teams')}

        for sf_uid, sf_ups in sf_user_perms.items():
            share_type = SharePermissions.SharePermissionsType.SF_USER
            load_user_permissions(sf_ups, sf_uid, share_type)
        for sf_uid, teams in team_perms.items():
            load_team_permissions(teams, sf_uid)

        apply_role_restrictions()



#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Commander
# Copyright 2023 Keeper Security Inc.
# Contact: ops@keepersecurity.com
#

import argparse
import fnmatch
import itertools
import json
import logging
import re
from typing import Union, Iterable, Optional, List

from .base import GroupCommand, Command, FolderMixin, try_resolve_path, report_output_parser, dump_report_data
from .. import vault, utils, crypto, api
from ..params import KeeperParams
from ..error import CommandError
from ..proto import record_pb2
from ..api import get_records_update_request
from ..security_audit import attach_security_data

kf_parse = argparse.ArgumentParser(add_help=False)
kf_parse.add_argument('-r', '--recursive', dest='recursive', action='store_true',
                           help='Traverse recursively through subfolders')
kf_parse.add_argument('paths', nargs='+', type=str, help='folder or record path or UID')


kf_list_parser = argparse.ArgumentParser(prog='keeper-fill list', parents=[kf_parse, report_output_parser])
kf_list_parser.add_argument('-v', '--verbose', dest='verbose', action='store_true', help='Do not truncate long names')


kf_set_parser = argparse.ArgumentParser(prog='keeper-fill set', parents=[kf_parse])
kf_set_parser.add_argument('--auto-fill', dest='auto_fill', action='store', choices=['on', 'off', 'none'],
                           help='Auto Fill')
kf_set_parser.add_argument('--auto-submit', dest='auto_submit', action='store', choices=['on', 'off', 'none'],
                           help='Auto Submit')


class KeeperFillCommand(GroupCommand):
    def __init__(self):
        super().__init__()
        self.register_command('list', KeeperFillListCommand(), 'Displays a list of KeeperFill values.')
        self.register_command('set', KeeperFillSetCommand(), 'Sets Keeper Fill settings.')
        self.default_verb = 'list'


class KeeperFillMixin:
    @staticmethod
    def resolve_records(params, **kwargs):   # type: (KeeperParams, ...) -> Iterable[str]
        recursive = kwargs.get('recursive') is True
        folders = []
        records = set()

        paths = kwargs.get('paths')
        if not paths:
            raise CommandError('', '"paths" parameter is required.')

        for path in paths:
            if path in params.folder_cache:
                folders.append(path)
            elif path in params.record_cache:
                records.add(path)
            else:
                rs = try_resolve_path(params, path)
                if rs is None:
                    raise CommandError('', f'Folder or record path {path} not found')

                folder, pattern = rs
                folder_uid = folder.uid or ''
                if len(pattern) == 0:
                    folders.append(folder_uid)
                else:
                    regex = re.compile(fnmatch.translate(pattern), re.IGNORECASE).match
                    if folder_uid in params.subfolder_record_cache:
                        for record_uid in params.subfolder_record_cache[folder_uid]:
                            if record_uid in params.record_cache:
                                rec = params.record_cache[record_uid]
                                if rec.get('version', 0) in (2, 3):
                                    r = vault.KeeperRecord.load(params, rec)
                                    if r and regex(r.title):
                                        records.add(r.record_uid)

        for folder_uid in folders:
            if recursive:
                records.update(FolderMixin.get_records_in_folder_tree(params, folder_uid))
            else:
                recs = params.subfolder_record_cache.get(folder_uid)
                if recs:
                    records.update(recs)

        return records

    @staticmethod
    def get_keeper_fill_data(params, record_uid):    # type: (KeeperParams, str) -> Optional[dict]
        if record_uid not in params.non_shared_data_cache:
            return
        if record_uid not in params.record_cache:
            return
        rd = params.record_cache[record_uid]
        if 'version' not in rd:
            return
        if rd['version'] not in (2, 3):
            return
        nsd = params.non_shared_data_cache[record_uid]
        if 'data_unencrypted' not in nsd:
            return
        try:
            return json.loads(nsd['data_unencrypted'])
        except Exception as e:
            logging.debug('Record UID: %s: Load NSD error: %s', record_uid, e)

    @staticmethod
    def get_record_url(record):      # type: (vault.KeeperRecord) -> Union[str, List[str], None]
        if isinstance(record, vault.PasswordRecord):
            if record.link:
                return record.link
        elif isinstance(record, vault.TypedRecord):
            url_fields = [x for x in itertools.chain(record.fields, record.custom) if x.type == 'url' and x.value]
            url = []
            for field in url_fields:
                url_str = field.get_default_value(str)
                if url_str:
                    url.append(url_str)
            if url:
                return url


class KeeperFillListCommand(Command, KeeperFillMixin):
    def get_parser(self):
        return kf_list_parser

    def execute(self, params, **kwargs):
        record_uids = self.resolve_records(params, **kwargs)

        verbose = kwargs.get('verbose') is True
        fmt = kwargs.get('format')
        if fmt != 'table':
            verbose = True

        table = []
        for record_uid in record_uids:
            data = self.get_keeper_fill_data(params, record_uid)
            if data is None:
                continue

            record = vault.KeeperRecord.load(params, record_uid)
            if not record:
                continue

            auto_fill_mode = data.get('auto_fill_mode')
            ext_auto_submit = data.get('ext_auto_submit')

            if isinstance(auto_fill_mode, str):
                if auto_fill_mode == 'always':
                    auto_fill_mode = True
                elif auto_fill_mode == 'never':
                    auto_fill_mode = False
                else:
                    auto_fill_mode = None

            url = self.get_record_url(record)

            if not url and auto_fill_mode is None and ext_auto_submit is None:
                continue

            title = record.title
            if not verbose:
                if len(record.title) > 32:
                    title = record.title[:30] + '...'
                if isinstance(url, str):
                    if len(url) > 32:
                        url = url[:30] + '...'

            table.append([record_uid, title, url, auto_fill_mode, ext_auto_submit])
        fmt = kwargs.get('format')
        headers = ['UID', 'Title', 'URL', 'Auto Fill', 'Auto Submit']
        return dump_report_data(table, headers, row_number=True, sort_by=1, fmt=fmt, filename=kwargs.get('output'))


class KeeperFillSetCommand(Command, KeeperFillMixin):
    def get_parser(self):
        return kf_set_parser

    def execute(self, params, **kwargs):
        auto_fill = kwargs.get('auto_fill')
        auto_submit = kwargs.get('auto_submit')
        if auto_fill is None and auto_submit is None:
            raise CommandError('', 'Nothing to set.')

        record_v3_updates = []   # type: List[record_pb2.RecordUpdate]
        record_v2_updates = []   # type: List[dict]
        record_uids = self.resolve_records(params, **kwargs)
        for record_uid in record_uids:
            data = self.get_keeper_fill_data(params, record_uid)
            if data is None:
                continue

            record = vault.KeeperRecord.load(params, record_uid)
            if not record:
                return

            url = self.get_record_url(record)
            if not url:
                continue

            auto_fill_mode = data.get('auto_fill_mode')
            ext_auto_submit = data.get('ext_auto_submit')

            if isinstance(auto_fill_mode, str):
                if auto_fill_mode == 'always':
                    auto_fill_mode = 'on'
                elif auto_fill_mode == 'never':
                    auto_fill_mode = 'off'
                else:
                    auto_fill_mode = 'none'
            else:
                auto_fill_mode = 'none'
            if ext_auto_submit is True:
                ext_auto_submit = 'on'
            elif ext_auto_submit is False:
                ext_auto_submit = 'off'
            else:
                ext_auto_submit = 'none'

            should_save = False
            if auto_fill is not None:
                if auto_fill_mode != auto_fill:
                    if auto_fill == 'on':
                        data['auto_fill_mode'] = 'always'
                    elif auto_fill == 'off':
                        data['auto_fill_mode'] = 'never'
                    elif auto_fill == 'none':
                        if 'auto_fill_mode' in data:
                            del data['auto_fill_mode']
                    should_save = True

            if auto_submit is not None:
                if ext_auto_submit != auto_submit:
                    if auto_submit == 'on':
                        data['ext_auto_submit'] = True
                    elif auto_submit == 'off':
                        data['ext_auto_submit'] = False
                    elif auto_submit == 'none':
                        if 'ext_auto_submit' in data:
                            del data['ext_auto_submit']
                    should_save = True

            if should_save:
                nsd = json.dumps(data).encode()
                if isinstance(record, vault.PasswordRecord):
                    nsd = crypto.encrypt_aes_v1(nsd, params.data_key)
                    ur = {
                        'record_uid': record_uid,
                        'revision': record.revision,
                        'non_shared_data': utils.base64_url_encode(nsd),
                        'version': 2,
                        'client_modified_time': utils.current_milli_time()
                    }
                    record_v2_updates.append(ur)
                elif isinstance(record, vault.TypedRecord):
                    ru = record_pb2.RecordUpdate()
                    ru.record_uid = utils.base64_url_decode(record_uid)
                    ru.client_modified_time = utils.current_milli_time()
                    ru.revision = record.revision
                    ru.non_shared_data = crypto.encrypt_aes_v2(nsd, params.data_key)
                    ru = attach_security_data(params, record, ru)
                    record_v3_updates.append(ru)

        if len(record_v3_updates) > 0 or len(record_v2_updates) > 0:
            params.sync_data = True

        while len(record_v3_updates) > 0:
            chunk = record_v3_updates[:900]
            record_v3_updates = record_v3_updates[900:]
            rq = get_records_update_request(params)
            rq.records.extend(chunk)

            rs = api.communicate_rest(params, rq, 'vault/records_update')

        while len(record_v2_updates):
            chunk = record_v2_updates[:90]
            record_v2_updates = record_v2_updates[90:]
            rq = {
                'command': 'record_update',
                'update_records': chunk
            }
            rs = api.communicate(params, rq)



#  _  __  
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|            
#
# Keeper Commander 
# Copyright 2021 Keeper Security Inc.
# Contact: ops@keepersecurity.com
#

import getpass
import logging
import os
import threading
from typing import Optional

from fido2.client import ClientError, DefaultClientDataCollector, UserInteraction, WebAuthnClient

from fido2.ctap import CtapError
from fido2.hid import CtapHidDevice
from fido2.webauthn import (PublicKeyCredentialRequestOptions, AuthenticationResponse,
                            PublicKeyCredentialCreationOptions, RegistrationResponse,
                            UserVerificationRequirement)
from fido2.ctap2.pin import ClientPin, Ctap2
from prompt_toolkit import PromptSession

from .. import utils

def verify_rp_id_none(rp_id, origin):
    return True


prompt_session = None    # type: Optional[PromptSession]
if os.isatty(0) and os.isatty(1):
    prompt_session = PromptSession(multiline=False, complete_while_typing=False)


class CliInteraction(UserInteraction):
    def prompt_up(self):
        print("\nTouch the flashing Security key to authenticate or "
              "press Ctrl-C to resume with the primary two factor authentication...\n")

    def request_pin(self, permissions, rd_id):
        global prompt_session
        prompt = "Enter Security Key PIN: "
        if prompt_session:
            return prompt_session.prompt(prompt, is_password=True)
        else:
            return getpass.getpass(prompt)

    def request_uv(self, permissions, rd_id):
        print("User Verification required.")
        return True


def yubikey_register(request, force_pin=False):    # type: (dict, bool) -> Optional[RegistrationResponse]
    rq = request.copy()
    user_id = rq['user']['id']
    if isinstance(user_id, str):
        rq['user']['id'] = utils.base64_url_decode(user_id)
    challenge = rq['challenge']
    if isinstance(challenge, str):
        rq['challenge'] = utils.base64_url_decode(challenge)

    if force_pin:
        uv = rq['authenticatorSelection']['userVerification']
        if uv != UserVerificationRequirement.REQUIRED:
            rq['authenticatorSelection']['userVerification'] = UserVerificationRequirement.REQUIRED

    options = PublicKeyCredentialCreationOptions.from_dict(rq)   # type: PublicKeyCredentialCreationOptions
    origin = options.extensions.get('appidExclude') or options.rp.id

    client = None   # type: Optional[WebAuthnClient]
    data_collector = DefaultClientDataCollector(origin, verify=verify_rp_id_none)
    if os.name == 'nt':
        from fido2.client.windows import WindowsClient
        client = WindowsClient(client_data_collector=data_collector)
    else:
        dev = next(CtapHidDevice.list_devices(), None)
        if not dev:
            logging.warning("No Security Key detected")
            return

        from fido2.client import Fido2Client
        fido_client = Fido2Client(dev, client_data_collector=data_collector, user_interaction=CliInteraction())
        uv_configured = any(fido_client.info.options.get(k) for k in ("uv", "clientPin", "bioEnroll"))
        uv = options.authenticator_selection.user_verification
        if uv == UserVerificationRequirement.REQUIRED:
            if not uv_configured:
                print('\nSecret Key PIN is required')
                answer = input('Do you want to setup PIN code for your Secret Key? (y/n): ')
                if answer not in ('y', 'Y'):
                    return
                prompt1 = '       PIN Code: '
                prompt2 = ' PIN Code Again: '
                if prompt_session:
                    pin1 = prompt_session.prompt(prompt1, is_password=True)
                else:
                    pin1 = getpass.getpass(prompt1)
                if not pin1:
                    raise Exception('PIN is required')
                if prompt_session:
                    pin2 = prompt_session.prompt(prompt2, is_password=True)
                else:
                    pin2 = getpass.getpass(prompt2)
                if not pin2:
                    raise Exception('PIN is required')
                if pin1 != pin2:
                    raise Exception('PINs do not match')
                client_pin = ClientPin(Ctap2(dev))
                client_pin.set_pin(pin1)
        elif uv == UserVerificationRequirement.PREFERRED:
            if not uv_configured:
                rq['authenticatorSelection']['userVerification'] = UserVerificationRequirement.DISCOURAGED
                options = PublicKeyCredentialCreationOptions.from_dict(rq)
        client = fido_client

    evt = threading.Event()
    try:
        return client.make_credential(options, event=evt)
    except ClientError as err:
        if isinstance(err.cause, CtapError):
            if err.cause.code == CtapError.ERR.PIN_INVALID:
                raise Exception('PIN is invalid')
            elif err.cause.code == CtapError.ERR.PIN_AUTH_BLOCKED:
                raise Exception('PIN is blocked')
        elif isinstance(err.cause, str):
            if err.code == ClientError.ERR.CONFIGURATION_UNSUPPORTED:
                raise Exception('Security key user verification (PIN or Biometric) is not configured')
        raise err
    except Exception as e:
        raise e
    finally:
        evt.set()


def yubikey_authenticate(request):  # type: (dict) -> Optional[AuthenticationResponse]
    if 'publicKeyCredentialRequestOptions' not in request:
        return

    options = request['publicKeyCredentialRequestOptions'].copy()
    origin = ''
    if 'extensions' in options:
        extensions = options['extensions']
        origin = extensions.get('appid') or ''
        if 'largeBlob' not in options['extensions']:
            options['extensions']['largeBlob'] = {'read': None}

    credentials = options.get('allowCredentials') or []
    for c in credentials:
        if isinstance(c.get('id'), str):
            c['id'] = utils.base64_url_decode(c['id'])

    challenge = options['challenge']
    if isinstance(challenge, str):
        options['challenge'] = utils.base64_url_decode(challenge)

    client = None   # type: Optional[WebAuthnClient]
    data_collector = DefaultClientDataCollector(origin, verify=verify_rp_id_none)
    if os.name == 'nt':
        from fido2.client.windows import WindowsClient
        client = WindowsClient(client_data_collector=data_collector)
    else:
        dev = next(CtapHidDevice.list_devices(), None)
        if not dev:
            logging.warning("No Security Key detected")
            return

        from fido2.client import Fido2Client
        fido_client = Fido2Client(dev, client_data_collector=data_collector, user_interaction=CliInteraction())

        uv_configured = any(fido_client.info.options.get(k) for k in ("uv", "clientPin", "bioEnroll"))
        if not uv_configured:
            uv = options['userVerification']
            if uv == UserVerificationRequirement.PREFERRED:
                options['userVerification'] = UserVerificationRequirement.DISCOURAGED
        client = fido_client

    if client is None:
        return

    evt = threading.Event()
    try:
        try:
            rq_options = PublicKeyCredentialRequestOptions.from_dict(options)  # type: PublicKeyCredentialRequestOptions
            rs = client.get_assertion(rq_options, event=evt)
            return rs.get_response(0)
        except ClientError as err:
            if isinstance(err.cause, CtapError):
                if err.cause.code == CtapError.ERR.NO_CREDENTIALS:
                    print('\n\nKeeper Security stopped supporting U2F security keys starting February 2022.\n'
                          'If you registered your security key prior to this date please re-register it within the Web Vault.\n'
                          'For information on using security keys with Keeper see the documentation: \n'
                          'https://docs.keeper.io/enterprise-guide/two-factor-authentication#security-keys-fido-webauthn\n'
                          'Commander will use the fallback security key authentication method.\n\n'
                          'To use your Yubikey with Commander, please touch the flashing Security key one more time.\n')
                    options['rpId'] = origin
                    rq_options = PublicKeyCredentialRequestOptions.from_dict(options)
                    rs = client.get_assertion(rq_options, event=evt)
                    return rs.get_response(0)
                elif err.cause.code == CtapError.ERR.PIN_INVALID:
                    raise Exception('PIN is invalid')
                elif err.cause.code == CtapError.ERR.PIN_AUTH_BLOCKED:
                    raise Exception('PIN is blocked')
            elif isinstance(err.cause, str):
                if err.code == ClientError.ERR.CONFIGURATION_UNSUPPORTED:
                    raise Exception('Security key user verification (PIN or Biometric) is not configured')
            raise err

    except KeyboardInterrupt:
        return
    except Exception as e:
        raise e
    finally:
        evt.set()



#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Commander
# Copyright 2022 Keeper Security Inc.
# Contact: ops@keepersecurity.com
#

"""Commander Plugin for SFTP
   Dependencies:
       pip install paramiko
"""

import logging
import stat

from typing import Optional

from paramiko.sftp_client import SFTPClient
from paramiko import Transport

from ...commands.base import RecordMixin
from .. import rsync

logging.getLogger("paramiko").setLevel(logging.WARNING)


class SFtpPlugin(rsync.RSyncPluginBase):
    def __init__(self):
        self._client = None      # type: Optional[SFTPClient]

    def connect(self, record):
        self.disconnect()

        host = RecordMixin.get_record_field(record, 'host')
        if host:
            host, _, port = host.partition(':')
        else:
            port = ''
        if not port:
            port = RecordMixin.get_record_field(record, 'port')
        if port:
            port = int(port)
        else:
            port = None
        if not host:
            raise ValueError(f'SFTP plugin: SFTP server (\"host\") is not set.')
        transport = Transport((host, port))

        logging.info('SFTP: connecting to %s', host)

        username = RecordMixin.get_record_field(record, 'login')
        if not username:
            raise ValueError(f'SFTP plugin: (\"login\") is not set.')
        password = RecordMixin.get_record_field(record, 'password')
        if not password:
            raise ValueError(f'SFTP plugin: (\"password\") is not set.')

        transport.connect(None, username, password)
        client = SFTPClient.from_transport(transport)
        self._client = client

    def disconnect(self):
        if self._client:
            self._client.close()
            self._client = None

    def get_entries(self, root_dir=None):
        if not self._client:
            raise Exception('Not connected')
        if root_dir:
            if not root_dir.endswith('/'):
                root_dir += '/'
        else:
            root_dir = '/'

        dirs = [root_dir or '/']
        pos = 0
        while pos < len(dirs):
            cwd = dirs[pos]
            if cwd.endswith('/'):
                cwd = cwd[0:-1]
            pos += 1
            for entry in self._client.listdir_attr(path=cwd):
                if entry.filename.startswith('.'):
                    continue
                name = f'{cwd}/{entry.filename}'
                if stat.S_ISDIR(entry.st_mode):
                    dirs.append(name)
                    pass
                elif stat.S_ISREG(entry.st_mode):
                    e = rsync.RSyncFileEntry(name[len(root_dir):])
                    e.size = entry.st_size
                    e.last_modified = int(entry.st_mtime)
                    e.full_path = name
                    yield e

    def get_entry_stream(self, entry):
        if not self._client:
            raise Exception('Not connected')
        return self._client.open(entry.full_path)



import logging

from keepercommander.auth import login_steps
from keepercommander import api
from keepercommander.__main__ import get_params_from_config


class CustomUi(login_steps.LoginUi):
    def on_device_approval(self, step):
        try:
            step.send_push(login_steps.DeviceApprovalChannel.Email)
        except Exception as e:
            logging.error(e)
        verification_code = wait_for_verification_code()
        step.send_code(login_steps.DeviceApprovalChannel.Email, verification_code)

    def on_two_factor(self, step):
        raise NotImplementedError()

    def on_password(self, step):
        password = wait_for_password()
        try:
            step.verify_password(password)
        except Exception as e:
            logging.error(e)

    def on_sso_redirect(self, step):
        raise NotImplementedError()

    def on_sso_data_key(self, step):
        raise NotImplementedError()


my_params = get_params_from_config('config.json')
while not my_params.user:
    my_params.user = input('User(Email): ')

ui = CustomUi()
api.login(my_params, login_ui=ui)

if not my_params.session_token:
    exit(1)


from __future__ import annotations
import logging
import os
from .vertex import DAGVertex
from .types import DAGData, EdgeType, RefType, Ref, DataPayload
from .crypto import encrypt_aes, decrypt_aes, generate_uid_str, bytes_to_str, str_to_bytes, urlsafe_str_to_bytes
from .exceptions import (DAGConfirmException, DAGPathException, DAGVertexAlreadyExistsException, DAGKeyException,
                         DAGVertexException, DAGCorruptException, DAGDataException)
from .utils import value_to_boolean
from .__version__ import __version__
import json
import importlib
from typing import Optional, Union, List, Any, TYPE_CHECKING

if TYPE_CHECKING:
    from .connection import ConnectionBase
    Content = Union[str, bytes, dict]
    QueryValue = Union[list, dict, str, float, int, bool]
    Logger = Union[logging.RootLogger, logging.Logger]


class DAG:

    # Debug level. Increase to get finer debug messages.
    DEBUG_LEVEL = 0

    UID_KEY_BYTES_SIZE = 16
    UID_KEY_STR_SIZE = 22

    # For the dot output, enum to text.
    EDGE_LABEL = {
        EdgeType.DATA: "DATA",
        EdgeType.KEY: "KEY",
        EdgeType.LINK: "LINK",
        EdgeType.ACL: "ACL",
        EdgeType.DELETION: "DELETION",
    }

    def __init__(self, conn: ConnectionBase, record: Optional[object] = None, key_bytes: Optional[bytes] = None,
                 name: Optional[str] = None, graph_id: int = 0, auto_save: bool = False,
                 history_level: int = 0, logger: Optional[Any] = None, debug_level: int = 0, is_dev: bool = False,
                 vertex_type: RefType = RefType.PAM_NETWORK, decrypt: bool = True, fail_on_corrupt: bool = True,
                 data_requires_encryption: bool = False, log_prefix: str = "GraphSync",
                 save_batch_count: Optional[int] = None, agent: Optional[str] = None):

        """
        Create a GraphSync instance.

        :param conn: Connection instance
        :param record: If set, the key bytes will use the key bytes in the record. Overrides key_bytes.
        :param key_bytes:  If set, these key bytes will be used.
        :param name: Optional name for the graph.
        :param graph_id: Graph ID sets which graph to load for the graph.
        :param auto_save: Automatically save when modifications are performed. Default is False.
        :param history_level: How much edge history to keep in memory. Default is 0, no history.
        :param logger: Python logger instance to use for logging.
        :param debug_level: Debug level; the higher the number will result in more debug information.
        :param is_dev: Is the code running in a development environment?
        :param vertex_type: The default vertex/ref type for the root vertex, if auto creating.
        :param decrypt: Decrypt the graph; Default is TRUE
        :param fail_on_corrupt: If unable to decrypt encrypted data, fail out.
        :param data_requires_encryption: Data edges are already encrypted. Default is False.
        :param log_prefix: Text prepended to the log messages. Handy if dealing with multiple graphs
        :param save_batch_count: The number of edges to save at one time.
        :param agent: User Agent to send with web service requests.
        :return: Instance of GraphSync
        """

        if logger is None:
            logger = logging.getLogger()
        self.logger = logger
        if debug_level is None:
            debug_level = int(os.environ.get("GS_DEBUG_LEVEL", os.environ.get("DAG_DEBUG_LEVEL", 0)))

        self.debug_level = debug_level
        self.log_prefix = log_prefix

        if save_batch_count is None or save_batch_count <= 0:
            save_batch_count = 0
        self.save_batch_count = save_batch_count
        self.debug(f"save batch count is set to {self.save_batch_count}")

        self.vertex_type = vertex_type

        self.data_requires_encryption = data_requires_encryption
        self.decrypt = decrypt
        self.fail_on_corrupt = fail_on_corrupt

        gs_is_dev = os.environ.get("GS_IS_DEV", os.environ.get("DAG_IS_DEV"))
        if gs_is_dev is not None:
            is_dev = value_to_boolean(gs_is_dev)
        self.is_dev = is_dev
        if self.is_dev is True:
            self.debug("GraphSync is running in a development environment, vertex names will be included.")

        # If the record is passed in, use the UID and key bytes from the record.
        self.uid = None
        if record is not None:
            self.uid = conn.get_record_uid(record)
            key_bytes = conn.get_key_bytes(record)

        self.key = key_bytes

        if key_bytes is None:
            raise ValueError("Either the record or the key_bytes needs to be passed.")

        # If the UID is blank, use the key bytes to generate a UID
        if self.uid is None:
            self.uid = generate_uid_str(key_bytes[:16])

        # Graph ID allow you to select which graph to load. The default is 0, which will load all graph for the UID
        self.graph_id = graph_id

        self.debug(f"{self.log_prefix} key {self.key}", level=1)
        self.debug(f"{self.log_prefix} UID {self.uid}", level=1)
        self.debug(f"{self.log_prefix} UID HEX {urlsafe_str_to_bytes(self.uid).hex()}", level=1)

        if name is None:
            name = f"{self.log_prefix} ROOT"
        self.name = name

        # The order of the vertices is important.
        # The order creates the history.
        # The web service will order edge by their edge_id
        # Store in and array.
        # The lookup table to make UID to DAGVertex easier.
        # The integer is the index into the array.
        self._vertices = []  # type: List[DAGVertex]
        self._uid_lookup = {}  # type: dict[str, int]

        # This is like the batch
        self.origin_uid = generate_uid_str()

        # If True, any addition or changes will automatically be saved.
        self.auto_save = auto_save

        # To auto save, both allow_auto_save and auto_save needs to be True.
        # If the graph has not been saved before and the root vertex has not been connected,
        #   we want to disable auto save.
        self._allow_auto_save = False

        # For big changes, we need a confirmation to save.
        self.need_save_confirm = False

        # The last sync point after save.
        self.last_sync_point = 0

        # Amount of history to keep.
        # The default is 0, which will keep all history.
        # Setting to 1 will only keep the latest edges.
        # Settings to 2 will keep the latest and prior edges.
        # And so on.
        self.history_level = history_level

        # If data was corrupt in the graph, the vertex UID will appear in this list.
        self.corrupt_uids = []

        self.conn = conn

        self.agent = f"keeper-dag/{__version__}"
        if agent is not None:
            self.agent += "; " + agent

    def debug(self, msg: str, level: int = 0):
        """
        Debug with granularity level.

        If the debug level is greater or equal to the level on the message, the message will be displayed.

        :param msg: Text debug message
        :param level: Debug level of message
        :return:
        """

        if self.debug_level >= level:

            msg = f"{self.log_prefix}: {msg}"

            if self.logger is not None:
                self.logger.debug(msg)
            else:
                logging.debug(msg)

    def __str__(self):
        ret = f"GraphSync {self.uid}\n"
        ret += f"  python instance id: {id(self)}\n"
        ret += f"  name: {self.name}\n"
        ret += f"  key: {self.key}\n"
        ret += f"  vertices:\n"
        for v in self.all_vertices:
            ret += f"    * {v.uid}, Keys: {v.keychain}, Active: {v.active}\n"
            for e in v.edges:
                if e.edge_type == EdgeType.DATA:
                    ret += f"      + has a DATA edge"
                    if e.content is not None:
                        ret += ", has content"
                else:
                    ret += f"      + belongs to {e.head_uid}, {DAG.EDGE_LABEL.get(e.edge_type)}, {e.content}"
                ret += "\n"

        return ret

    @property
    def is_corrupt(self):
        return len(self.corrupt_uids) > 0

    @property
    def allow_auto_save(self) -> bool:
        """
        Return the flag indicating if auto save is allowed.
        :return:
        """

        return self._allow_auto_save

    @allow_auto_save.setter
    def allow_auto_save(self, value: bool):
        """
        Set the ability to auto save.
        :param value: True enables, False disables.
        :return:
        """

        if value is True:
            self.debug("ability to auto save has been ENABLED", level=2)
        else:
            self.debug("ability to auto save has been DISABLED", level=2)

        self._allow_auto_save = value

    @property
    def origin_ref(self) -> Ref:

        """
        Return an instance of the origin reference.
        :return:
        """

        return Ref(
            type=RefType.DEVICE,
            value=self.origin_uid,
            name=self.name if self.is_dev is True else None
        )

    @property
    def has_graph(self) -> bool:
        """
        Do we have any graph items?

        :return: True if there are vertices. False if no vertices.
        """

        return len(self._vertices) > 0

    @property
    def vertices(self) -> List[DAGVertex]:
        """
        Get all active vertices

        :return: List of DAGVertex instance
        """

        return [
            vertex
            for vertex in self._vertices
            if vertex.active is True
            ]

    @property
    def all_vertices(self) -> List[DAGVertex]:
        """
        Get all vertices
        :return: List of DAGVertex instance
        """

        return self._vertices

    def get_vertex(self, key) -> Optional[DAGVertex]:

        """
        Get a single vertex.

        The key can be either a UID, path or name.

        The UID is most reliable since there can only be one per graph.

        The path is second reliable if it is set by the user.
        It will find an edge with the path, the vertex that is the edge's tail is returned.
        There is no unique constraint for the path.
        You can have duplicates.

        The name is third, and not reliable.
        The name only exists when the graph is created.
        If loaded, the name will be None.

        :param key: A UID, path item, or name of a vertex.
        :return: DAGVertex instance, if it exists.
        """

        if key is None:
            return None

        # Is the key a UID? If so, return the vertex from the lookup.
        if key in self._uid_lookup:
            index = self._uid_lookup[key]
            return self._vertices[index]

        # Is the key a path?
        # We also want to include any deleted edges.
        vertices = self.get_vertices_by_path_value(key, inc_deleted=True)
        if len(vertices) > 0:
            if len(vertices) > 1:
                raise DAGPathException("Cannot get vertex using the path. Found multiple vertex that use the path.")
            return vertices[0]

        # Is the key a name? This is a last resort.
        for vertex in vertices:
            if vertex.name == key:
                return vertex

        return None

    @property
    def get_root(self) -> Optional[DAGVertex]:
        """
        Get the root vertex

        If the root vertex does not exist, it will create the vertex with a ref type of PAM_NETWORK.

        :return:
        """
        root = self.get_vertex(self.uid)
        if root is None:
            root = self.add_vertex(uid=self.uid, name=self.name, vertex_type=self.vertex_type)
        return root

    def vertex_exists(self, key: str) -> bool:
        """
        Check if a vertex identified by the key exists.
        :param key: UID, path, or name
        :return:
        """

        return self.get_vertex(key) is not None

    def get_vertices_by_path_value(self, path: str, inc_deleted: bool = False) -> List[DAGVertex]:
        """
        Find all vertices that have an edge that match the path
        :param path: A string path value. This is a path to walk, just the value.
        :param inc_deleted: Include deleted edges.
        :return: List of DAGVertex
        """
        results = []
        if inc_deleted is True:
            vertices = self.all_vertices
        else:
            vertices = self.vertices

        for vertex in vertices:
            for edge in vertex.edges:
                if edge.path == path:
                    results.append(vertex)
        return results

    def _sync(self, sync_point: int = 0) -> (List[DAGData], int):

        # The web service will send 500 items, if there is more the 'has_more' flag is set to True.
        has_more = True

        # Make the web service call to set all the data
        all_data = []
        while has_more is True:
            # Load a page worth of items
            resp = self.conn.sync(
                stream_id=self.uid,
                sync_point=sync_point,
                graph_id=self.graph_id,
                agent=self.agent
            )
            if resp.syncPoint == 0:
                return all_data, 0

            all_data += resp.data

            # The server will tell us if there is more data to get.
            has_more = resp.hasMore

            # The sync_point will indicate where we need to start the sync from. Think syncPoint > value
            sync_point = resp.syncPoint

        return all_data, sync_point

    def _load(self, sync_point: int = 0):

        """
        Load the DAG

        This will clear the existing graph.
        It will make web services calls to get the fresh graph, which will return a list of edges.
        With the list of edges, it will create vertices and connect them with the edges.
        The content of the edges will remain encrypted. The 'encrypted' flag is set to True.
        We need the entire graph structure before decrypting.

        We don't have to worry about keys at this point. We are just trying to get structure
        and content in the right place. Nothing is decrypted here.

        :param sync_point: Where to load
        """

        # Clear the existing vertices.
        self._vertices = []  # type: List[DAGVertex]
        self._uid_lookup = {}  # type: dict[str, int]

        self.debug("# SYNC THE GRAPH ##################################################################", level=1)

        # Make the web service call to set all the data
        all_data, sync_point = self._sync(sync_point=sync_point)

        self.debug("  PROCESS the non-DATA edges", level=2)

        # Process the non-DATA edges
        for data in all_data:

            # Skip all the DATA edge
            edge_type = EdgeType.find_enum(data.type)
            if edge_type == EdgeType.DATA:
                continue

            # The ref the tail. It connects to stored in the vertex.
            tail_uid = data.ref.get("value")

            # The parentRef is the head. It's the arrowhead on the edge. For DATA edges, it will be None.
            head_uid = None
            if data.parentRef is not None:
                head_uid = data.parentRef.get("value")

            self.debug(f"  * edge {edge_type}, tail {tail_uid} to head {head_uid}", level=3)

            # We want to store this edge in the Vertex with the same value/UID as the ref.
            if self.vertex_exists(tail_uid) is False:
                self.debug(f"    * tail vertex {tail_uid} does not exists. create.", level=3)
                self.add_vertex(
                    uid=tail_uid,
                    name=data.ref.get("name"),

                    # This will be 0/GENERAL right now. We do the lookup just in case things will change in the
                    # future.
                    vertex_type=RefType.find_enum(data.ref.get("type"))
                )

            # Get the tail vertex.
            tail = self.get_vertex(tail_uid)

            # This most likely is a DELETION edge of a DATA edge.
            # Set the head to be the same as the tail.
            if head_uid is None:
                head_uid = tail_uid

            # If the head vertex doesn't exist, we need to create.
            if self.vertex_exists(head_uid) is False:
                self.debug(f"    * head vertex {head_uid} does not exists. create.", level=3)
                self.add_vertex(
                    uid=head_uid,
                    name=data.parentRef.get("name"),
                    vertex_type=RefType.GENERAL
                )
            # Get the head vertex, which will exist now.
            head = self.get_vertex(head_uid)
            self.debug(f"    * tail {tail_uid} belongs to {head_uid}, "
                       f"edge type {edge_type}", level=3)

            if edge_type == EdgeType.DELETION:
                tail.disconnect_from(head)
            else:
                if data.content is not None:
                    content = str_to_bytes(data.content)
                else:
                    content = None

                # ACL are decrypted, but it is base64 encode.
                # We need to deserialize the base64 to get the bytes.
                # We can't update an existing edges content after added.
                # if edge_type == EdgeType.ACL:
                #     content = str_to_bytes(content)

                # Connect this vertex to the head vertex. It belongs to that head vertex.
                tail.belongs_to(
                    vertex=head,
                    edge_type=edge_type,
                    # content is encrypted
                    content=content,
                    path=data.path,
                    modified=False,
                    from_load=True
                )

        self.debug("", level=2)
        self.debug("  PROCESS the DATA edges", level=2)

        # Process the DATA edges
        # We don't have to worry about vertex creation since they will all exist.
        for data in all_data:

            # Only process the data edges.
            edge_type = EdgeType.find_enum(data.type)
            if edge_type != EdgeType.DATA:
                continue

            # Get the tail vertex.
            tail_uid = data.ref.get("value")
            # We want to store this edge in the Vertex with the same value/UID as the ref.
            if self.vertex_exists(tail_uid) is False:
                self.debug(f"    * tail vertex {tail_uid} does not exists. create.", level=3)
                self.add_vertex(
                    uid=tail_uid,
                    name=data.ref.get("name"),

                    # This will be 0/GENERAL right now. We do the lookup just in case things will change in the
                    # future.
                    vertex_type=RefType.find_enum(data.ref.get("type"))
                )
            tail = self.get_vertex(tail_uid)

            self.debug(f"  * DATA edge belongs to {tail.uid}", level=3)
            tail.add_data(
                # content is encrypted
                content=data.content,
                path=data.path,
                modified=False,
                from_load=True,
            )

        self.debug("", level=1)

        return sync_point

    def _mark_deletion(self):

        """
        Mark vertices as deleted.

        Check each vertex to see if there is any non-DELETION edge connecting to another vertex.
        If there are no edges, then the vertex is flagged at deleted.

        This is done to prevent the edges from being connected to a deleted vertex.
        Also, to display deleted vertex in the DOT graph.
        :return:
        """

        self.debug("  CHECK dag vertices to see if they are active", level=1)
        for vertex in self.all_vertices:

            self.debug(f"check vertex {vertex.uid}", level=3)
            found_edge_to_another_vertex = False
            for edge in vertex.edges:
                # Skip the DELETION and DATA edges.
                if edge.edge_type == EdgeType.DELETION or edge.edge_type == EdgeType.DATA:
                    continue

                # Check if this edge has a matching DELETION edge.
                # If it does not, this vertex cannot be deleted.
                if edge.is_deleted is False:
                    found_edge_to_another_vertex = True
                    break

            # If the vertex belongs to no vertex, and it not the root, then flag it for deletion.
            if found_edge_to_another_vertex is False and vertex.uid != self.uid:
                self.debug(f"  * vertex is deleted", level=3)
                vertex.active = False

        self.debug("", level=1)

    def _decrypt_keychain(self):

        """
        Decrypt KEY/ACL edges

        Part one is to decrypt the KEY and ACL edges.
        To decrypt the edge, we need to walk up the edges until we can no longer.
        If we get the point where we can't walk up any farther, we need to use the record key bytes.
        While walking up, if we get to a keychain that has been decrypted, we return that keychain.
        As we walk back, we can decrypt any keychain that is still encrypted.
        The decrypt keychain is set in the vertex.
        """

        self.debug("  DECRYPT the dag KEY edges", level=1)

        def _get_keychain(v):
            self.debug(f"  * looking at {v.uid}", level=3)

            # If the vertex has a decrypted key, then return it.
            if v.has_decrypted_keys is True:
                self.debug("  found a decrypted keychain on vertex", level=3)
                return v.keychain

            # Else we need KEY/ACL edge and get the key from the vertex that this vertex belongs to
            found_key_edge = False
            for e in v.edges:
                if e.edge_type == EdgeType.KEY:

                    self.debug(f"    has edge that is a key, check head vertex {e.head_uid}", level=3)
                    head = self.get_vertex(e.head_uid)
                    keychain = _get_keychain(head)

                    # No need to check if keychain exists.
                    # At default, it should contain the record bytes if no KEY/ACL edges existed for a vertex.

                    self.debug(f"  * decrypt {v.uid} with keys {keychain}", level=3)
                    was_able_to_decrypt = False

                    # Try the keys in the keychain. One should be able to decrypt the content.
                    for key in keychain:
                        try:
                            # The edge will contain a single key.
                            # Adding a key to
                            self.debug(f"    decrypt with key {key}", level=3)
                            content = decrypt_aes(e.content, key)
                            self.debug(f"    content {content}", level=3)
                            v.add_to_keychain(content)
                            self.debug(f"  * vertex {v.uid} keychain is {v.keychain}", level=3)
                            was_able_to_decrypt = True
                            found_key_edge = True
                            break
                        except (Exception,):
                            self.debug(f"      !! this is not the key", level=3)

                    if was_able_to_decrypt is False:

                        # Flag that the edge is corrupt, flag that the vertex keychain is corrupt,
                        #   and store vertex UID/tail UID.
                        # If we fail on corrupt keys, then raise exceptions.
                        e.corrupt = True
                        v.corrupt = True
                        self.corrupt_uids.append(v.uid)
                        if self.fail_on_corrupt is True:
                            raise DAGKeyException(f"Could not decrypt vertex {v.uid} keychain for edge path {e.path}")
                        return []

            if found_key_edge is True:
                return v.keychain
            else:
                self.debug("  * using record bytes", level=3)
                return [self.key]

        for vertex in self.all_vertices:
            if vertex.has_key is False:
                continue
            self.debug(f"vertex {vertex.uid}, {vertex.has_key}, {vertex.has_decrypted_keys}", level=3)
            vertex.keychain = _get_keychain(vertex)
            self.debug(f"  setting keychain to {vertex.keychain}", level=3)

        self.debug("", level=1)

    def _decrypt_data(self):

        """
        Decrypt DATA edges

        At this point, all the vertex should have an encrypted key.
        This key is used to decrypt the DATA edge's content.
        Walk each vertex and decrypt the DATA edge if there is a DATA edge.
        """

        self.debug("  DECRYPT the dag data", level=1)
        for vertex in self.all_vertices:
            if vertex.has_data is False:
                continue
            self.debug(f"vertex {vertex.uid}, {vertex.keychain}", level=3)

            for edge in vertex.edges:
                if edge.edge_type != EdgeType.DATA:
                    continue

                # If the vertex/KEY edge that tail is this vertex is corrupt, we cannot decrypt data.
                if vertex.corrupt is True:
                    self.logger.error(f"the key for the DATA edge is corrupt for vertex {vertex.uid}; "
                                      "cannot decrypt data.")
                    continue

                content = edge.content
                if isinstance(content, bytes) is True:
                    raise ValueError("The content has already been decrypted.")

                self.debug(f"  * enc safe content {content}", level=3)
                if isinstance(content, str):
                    content = str_to_bytes(content)
                self.debug(f"  * enc {content}, enc key {vertex.keychain}", level=3)
                able_to_decrypt = False

                keychain = vertex.keychain

                # Try the keys in the keychain. One should be able to decrypt the content.
                for key in keychain:
                    try:
                        edge.content = decrypt_aes(content, key)
                        able_to_decrypt = True
                        self.debug(f"  * content {edge.content}", level=3)
                        break
                    except (Exception,):
                        self.debug(f"      !! this is not the key", level=3)

                if able_to_decrypt is False:

                    # If the DATA edge requires encryption, throw error if we cannot decrypt.
                    if self.data_requires_encryption is True:
                        self.corrupt_uids.append(vertex.uid)
                        raise DAGDataException(f"The data edge {vertex.uid} could not be decrypted.")

                    edge.content = content
                    edge.needs_encryption = False
                    self.debug(f"  * edge is not encrypted or key is incorrect.")

        self.debug("", level=1)

    def _flag_as_not_modified(self):

        """
        Flag all edges a not modified.

        :return:
        """

        for vertex in self.all_vertices:
            for edge in vertex.edges:
                edge.modified = False

    def load(self, sync_point: int = 0) -> int:

        """
        Load data from the graph.

        The first step is to recreate the structure of the graph.
        The second step is mark vertex as deleted.
        The third step is to decrypt the KEY/ACL/DATA edges.
        Forth is to flag all edges as not modified.

        :return: The sync point of the graph stream
        """

        # During the load, turn off auto save
        self.allow_auto_save = False

        self.debug("== LOAD DAG ========================================================================", level=2)
        sync_point = self._load(sync_point)
        self.debug(f"sync point is {sync_point}")
        self._mark_deletion()
        if self.decrypt is True:
            self._decrypt_keychain()
            self._decrypt_data()
        else:
            self.logger.info("the DAG has not been decrypted, the decrypt flag was get to False")
        self._flag_as_not_modified()
        self.debug("====================================================================================", level=2)

        # We have loaded the grpah, enable the ability to use auto save.
        self.allow_auto_save = True

        self.last_sync_point = sync_point

        return sync_point

    def _make_delta_graph(self, duplicate_data: bool = True):

        self.debug("DELTA GRAPH", level=3)
        modified_vertices = []
        for vertex in self.all_vertices:
            found_modification = False
            for edge in vertex.edges:
                if edge.modified is True:
                    found_modification = True
                    break
            if found_modification is True:
                modified_vertices.append(vertex)
        if len(modified_vertices) == 0:
            self.debug("nothing has been modified")
            return

        self.debug(f"has {len(modified_vertices)} vertices", level=3)

        def _flag(vertex: DAGVertex):

            self.debug(f"check vertex {vertex.uid}", level=3)
            if vertex.uid == self.uid:
                self.debug(f"  FOUND ROOT", level=3)
                return True

            # Check if we have any of these edges in this order.
            found_path = False
            for edge_type in [EdgeType.KEY, EdgeType.ACL, EdgeType.LINK]:
                seen = {}
                for edge in vertex.edges:
                    self.debug(f"  checking {edge.edge_type}, {vertex.uid} to {edge.head_uid}", level=3)
                    is_deletion = None
                    if edge.edge_type == edge_type:
                        self.debug(f"    found {edge_type}", level=3)
                        next_vertex = self.get_vertex(edge.head_uid)

                        if is_deletion is None:
                            # If the most recent edge a DELETION edge?
                            version, highest_edge = vertex.get_highest_edge_version(next_vertex.uid)
                            is_deletion = highest_edge.edge_type == EdgeType.DELETION
                            if is_deletion is True:
                                self.debug(f"    highest deletion edge. will not mark any edges as modified",
                                           level=3)

                        found_path = _flag(next_vertex)
                        if found_path is True and seen.get(edge.head_uid) is None:
                            self.debug(f"  setting {vertex.uid}, {edge_type} active", level=3)
                            if is_deletion is False:
                                edge.modified = True
                                seen[edge.head_uid] = True
                    else:
                        self.debug(f"    edge is not {edge_type}", level=3)

                if found_path is True:
                    break

            # If we found a path, we may need to duplicate the DATA edge.
            if found_path is True and duplicate_data is True:
                for edge in vertex.edges:
                    if edge.edge_type == EdgeType.DATA:
                        edge.modified = True
                        break

            return found_path

        self.logger.debug("BEGIN delta graph edge detection")
        for modified_vertex in modified_vertices:
            _flag(modified_vertex)
        self.logger.debug("END delta graph edge detection")

    def save(self, confirm: bool = False, delta_graph: bool = False):

        """
        Save the graph

        We will not save if using the default graph.

        The save process will only save edges that have been flagged as modified, or are newly added.
        The process will get the edges from all vertices.
        The UID of the vertex is the tail UID of the edge.
        For DATA edges, the key (first key in the keychain) will be used for encryption.

        If the web service takes too long or hangs, the batch_count can be used to reduce the amount the web service
          needs to handle per request. If set to None or non-postivie value, it will not send in batches.

        :param confirm: Confirm save.
                        Only need this when deleting all vertices.
        :param delta_graph: Make a standalone graph from the modifications.
                            Use sync points to load this graph.

        :return:
        """

        self.debug("== SAVE GRAPH ========================================================================", level=2)

        if self.is_corrupt is True:
            self.logger.error(f"the graph is corrupt, there are problem UIDs: {','.join(self.corrupt_uids)}")
            raise DAGCorruptException(f"Cannot save. Graph steam uid {self.uid}, graph id {self.graph_id} "
                                      f"has corrupt vertices: {','.join(self.corrupt_uids)}")

        root_vertex = self.get_vertex(self.uid)
        if root_vertex is None:
            raise DAGVertexException("Cannot save. Could not find the root vertex.")

        if root_vertex.vertex_type != RefType.PAM_NETWORK and root_vertex.vertex_type != RefType.PAM_USER:
            raise DAGVertexException("Cannot save. Root vertex type needs to be PAM_NETWORK or PAM_USER.")

        # Do we need to the 'confirm' parameter set to True?
        # This is needed if the entire graph is being deleted.
        if self.need_save_confirm is True and confirm is False:
            raise DAGConfirmException("Cannot save. Confirmation is required.")
        self.need_save_confirm = False

        if delta_graph is True:
            self._make_delta_graph()

        data_list = []

        def _add_data(vertex):
            self.debug(f"processing vertex {vertex.uid}, key {vertex.key}, type {vertex.vertex_type}", level=3)
            # The vertex UID and edge tail UID
            uid = vertex.uid
            for edge in vertex.edges:
                self.debug(f"  * edge {edge.edge_type.value}, head {edge.head_uid}, tail {vertex.uid}", level=3)

                # If this edge is not modified, don't add to the data list to save.
                if edge.modified is False:
                    self.debug(f"    not modified, not saving.", level=3)
                    continue

                content = edge.content

                # If we are decrypting the edge data, then we want to encrypt it when we save.
                # Else, save the content as it is.
                if self.decrypt is True:
                    if edge.edge_type == EdgeType.DATA:
                        self.debug(f"    edge is data, encrypt data: {edge.needs_encryption}", level=3)
                        if isinstance(content, dict) is True:
                            content = json.dumps(content)
                        if isinstance(content, str) is True:
                            content = content.encode()

                        # If individual edges require encryption or all DATA edge require encryption, then encrypt
                        if edge.needs_encryption is True or self.data_requires_encryption is True:
                            self.debug(f"    content {edge.content}, enc key {vertex.key}", level=3)
                            content = encrypt_aes(content, vertex.key)
                            self.debug(f"    enc content {content}", level=3)

                        content = bytes_to_str(content)
                        self.debug(f"    enc safe content {content}", level=3)
                    elif edge.edge_type == EdgeType.KEY:
                        self.debug(f"    edge is key or acl, encrypt key", level=3)
                        head_vertex = self.get_vertex(edge.head_uid)
                        key = head_vertex.key
                        if key is None:
                            self.debug(f"     the edges head vertex {edge.head_uid} did not have a key. "
                                       "using root dag key.", level=3)
                            key = self.key
                        self.debug(f"    key {vertex.key}, enc key {key}", level=3)
                        content = bytes_to_str(encrypt_aes(vertex.key, key))
                    elif edge.edge_type == EdgeType.ACL:
                        content = bytes_to_str(edge.content)
                    else:
                        self.debug(f"    edge is {edge.edge_type}", level=3)

                parent_vertex = self.get_vertex(edge.head_uid)

                if content is not None and len(content) > 65_535:
                    self.debug(f"    !! vertex {vertex.uid} data edge is {len(content)}, over 64K.")
                    raise DAGDataException(f"vertex {vertex.uid} DATA edge is {len(content)} bytes. "
                                           "This is too large for the MySQL BLOB (64K).")

                data = DAGData(
                    type=edge.edge_type,
                    content=content,
                    # tail point at this vertex, so it uses this vertex's uid.
                    ref=Ref(
                        type=vertex.vertex_type,
                        value=uid,
                        name=vertex.name if self.is_dev is True else None
                    ),
                    # Head, the arrowhead, points at the vertex this vertex belongs to, the parent.
                    # Apparently, for DATA edges, the parentRef is allowed to be None.
                    # Doesn't hurt to send it.
                    parentRef=Ref(
                        type=parent_vertex.vertex_type,
                        value=edge.head_uid,
                        name=parent_vertex.name if self.is_dev is True else None
                    ),
                    path=edge.path
                )

                data_list.append(data)

                # Flag that this edge is no longer modified.
                edge.modified = False

        # Add the root vertex first
        _add_data(self.get_root)

        # Add the rest.
        # Only add is the skip_save is False.
        for v in self.all_vertices:
            if v.skip_save is False:
                if v.uid != self.uid:
                    _add_data(v)

        # Save the keys before the data.
        # This is done to make sure the web service can figure out the stream id.
        # By saving the keys before data, the structure of the graph is formed.
        if len(data_list) > 0:

            if self.debug_level >= 4:

                self.debug("EDGE LIST")
                self.debug("##############################################")
                for data in data_list:
                    self.debug(f"{data.ref.value} -> {data.parentRef.value} ({data.type})")
                self.debug("##############################################")

            self.debug(f"total list has {len(data_list)} items", level=0)
            self.debug(f"batch {self.save_batch_count} edges", level=0)

            batch_num = 0
            while len(data_list) > 0:

                # If using batch add, then take the first batch_count items.
                # Remove them from the data list
                if self.save_batch_count > 0:
                    batch_list = data_list[:self.save_batch_count]
                    data_list = data_list[self.save_batch_count:]

                # Else take everything and clear the data list (else infinite loop)
                else:
                    batch_list = data_list
                    data_list = []

                # Little sanity check
                if len(batch_list) == 0:
                    break

                self.debug(f"adding {len(batch_list)} edges, batch {batch_num}", level=0)
                payload = DataPayload(
                    origin=self.origin_ref,
                    dataList=batch_list,
                    graphId=self.graph_id
                )

                self.debug("PAYLOAD; batch {batch_num} =======================", level=5)
                self.debug(payload.model_dump_json(), level=5)
                self.debug("==================================================", level=5)

                try:
                    self.conn.add_data(payload, agent=self.agent)
                except Exception as err:
                    if value_to_boolean(os.environ.get("GS_SHOW_PAYLOAD_ERR", False)) is True:
                        self.logger.error(f"\n----\n{payload.model_dump_json(indent=4)}\n----\n")
                    raise err

                batch_num += 1

                # It's a POST that returns no data
        else:
            self.debug("data list was empty, not saving.", level=2)

        self.debug("====================================================================================", level=2)

    def do_auto_save(self):
        # If allow_auto_save is False, we will not allow auto saving.
        # On newly created graph, this will happen if the root vertex has not been connected.
        # The root vertex/disconnect edge head is needed to get a proper stream ID.
        if self.allow_auto_save is False:
            self.debug("cannot auto_save, allow_auto_save is False.", level=3)
            return
        if self.auto_save is True:
            self.debug("... dag auto saving", level=1)
            self.save()

    def add_vertex(self, name: Optional[str] = None, uid: Optional[str] = None,  keychain: Optional[List[bytes]] = None,
                   vertex_type: RefType = RefType.GENERAL) -> DAGVertex:

        """
        Add a vertex to the graph.

        :param name: Name for the vertex.
        :param uid: String unique identifier.
                    It's a 16bit hex value that is base64 encoded.
        :param keychain: List if key bytes to use for encryption/description. This is set by the load/save method.
        :param vertex_type: A RefType enumeration type. If blank, it will default to GENERAL.
        :return:
        """

        if name is None:
            name = uid

        vertex = DAGVertex(
            name=name,
            dag=self,
            uid=uid,
            keychain=keychain,
            vertex_type=vertex_type
        )
        if self.vertex_exists(vertex.uid) is True:
            raise DAGVertexAlreadyExistsException(f"Vertex {vertex.uid} already exists.")

        # Set the UID to array index lookup.
        # This is where the vertex will be in the vertices list.
        # Then append the vertex to the vertices list.
        self._uid_lookup[vertex.uid] = len(self._vertices)
        self._vertices.append(vertex)

        return vertex

    @property
    def is_modified(self) -> bool:
        for vertex in self.all_vertices:
            for edge in vertex.edges:
                if edge.modified is True:
                    return True
        return False

    @property
    def modified_edges(self):
        edges = []
        for vertex in self.all_vertices:
            for edge in vertex.edges:
                if edge.modified is True:
                    edges.append(edge)
        return edges

    def delete(self):
        """
        Delete the entire graph.

        This will delete all the vertex, which will delete all the edges.
        This will not automatically save.
        The save method will need to be called.
        The save will require the 'confirm' parameter to be set to True.
        :return:
        """
        for vertex in self.vertices:
            vertex.delete()
        self.need_save_confirm = True

    def _search(self, content: Any, value: QueryValue, ignore_case: bool = False):

        if isinstance(value, dict) is True:
            # If the object is not a dictionary, then it's not match
            if isinstance(content, dict) is False:
                return False
            for next_key, next_value in value.items():
                if next_key not in content:
                    return False
                if self._search(content=content[next_key],
                                value=next_value,
                                ignore_case=ignore_case) is False:
                    return False
            return True
        elif isinstance(value, list) is True:
            # If the object is not a dictionary, then it's not match
            for next_value in value:
                if self._search(content=content,
                                value=next_value,
                                ignore_case=ignore_case) is True:
                    return True
            return False
        else:
            content = str(content)
            value = str(value)
            if ignore_case is True:
                content = content.lower()
                value = value.lower()

            return value in content

    def search_content(self, query, ignore_case: bool = False):
        results = []
        for vertex in self.vertices:
            if vertex.has_data is False or vertex.active is False:
                continue
            content = vertex.content
            if isinstance(query, bytes) is True:
                if query == content:
                    results.append(vertex)
            elif isinstance(query, str) is True:
                try:
                    content = content.decode()
                    if query in content:
                        results.append(vertex)
                    continue

                except (Exception,):
                    pass
            elif isinstance(query, dict) is True:
                try:
                    content = content.decode()
                    content = json.loads(content)
                    search_result = self._search(content, value=query, ignore_case=ignore_case)
                    if search_result is True:
                        results.append(vertex)
                except (Exception,):
                    pass
            else:
                raise ValueError("Query is not an accepted type.")
        return results

    def walk_down_path(self, path: Union[str, List[str]]) -> Optional[DAGVertex]:

        """
        Walk the vertices using the path and return the vertex starting at root vertex.

        :param path: An array of path string, or string where the path is joined with a "/" (i.e., think URL)
        :return: DAGVertex is the path completes, None is failure.
        """

        self.debug("walking path starting at the root vertex", level=2)
        vertex = self.get_vertex(self.uid)
        return vertex.walk_down_path(path)

    def to_dot(self, graph_format: str = "svg", show_hex_uid: bool = False,
               show_version: bool = True, show_only_active: bool = False):

        """
        Generate a graphviz Gigraph in DOT format that is marked up.

        :param graph_format:
        :param show_hex_uid:
        :param show_version:
        :param show_only_active:
        :return:
        """

        try:
            mod = importlib.import_module("graphviz")
        except ImportError:
            raise Exception("Cannot to_dot(), graphviz module is not installed.")

        dot = getattr(mod, "Digraph")(comment=f"GraphSync for {self.name}", format=graph_format)
        dot.attr(rankdir='BT')

        for v in self._vertices:
            if show_only_active is True and v.active is False:
                continue
            if v.corrupt is False:
                fillcolor = "white"
                if v.active is False:
                    fillcolor = "grey"
                label = f"uid={v.uid}"
                if v.name is not None and v.name != v.uid:
                    label += f"\\nname={v.name}"
                if show_hex_uid is True:
                    label += f"\\nhex={urlsafe_str_to_bytes(v.uid).hex()}"
            else:
                fillcolor = "red"
                label = f"{v.uid} (CORRUPT)"

            dot.node(v.uid, label, fillcolor=fillcolor, style="filled")
            for edge in v.edges:

                if edge.corrupt is False:
                    color = "grey"
                    style = "solid"

                    # To reduce the number of edges, only show the active edges
                    if edge.active is True:
                        color = "black"
                        style = "bold"
                    elif show_only_active is True:
                        continue

                    # If the vertex is not active, gray out the DATA edge
                    if edge.active is False:
                        color = "grey"

                    if edge.edge_type == EdgeType.DELETION:
                        style = "dotted"

                    label = DAG.EDGE_LABEL.get(edge.edge_type)
                    if label is None:
                        label = "UNK"
                    if edge.path is not None and edge.path != "":
                        label += f"\\npath={edge.path}"
                    if show_version is True:
                        label += f"\\ne{edge.version}"
                    # tail, head (arrow side), label
                else:
                    color = "red"
                    style = "solid"
                    label = f"{DAG.EDGE_LABEL.get(edge.edge_type)} (CORRUPT)"

                dot.edge(v.uid, edge.head_uid, label, style=style, fontcolor=color, color=color)

        return dot

    def to_dot_raw(self, graph_format: str = "svg", sync_point: int = 0, rank_dir="BT"):

        """
        Generate a graphviz Gigraph in DOT format that is not (heavily) marked up.

        :param graph_format:
        :param sync_point:
        :param rank_dir:
        :return:
        """

        try:
            mod = importlib.import_module("graphviz")
        except ImportError:
            raise Exception("Cannot to_dot(), graphviz module is not installed.")

        dot = getattr(mod, "Digraph")(comment=f"GraphSync for {self.name}", format=graph_format)
        dot.attr(rankdir=rank_dir)

        all_data, sync_point = self._sync(sync_point=sync_point)

        for edge in all_data:
            edge_type = edge.type
            tail_uid = edge.ref.get("value")
            dot.node(tail_uid, tail_uid)
            if edge.parentRef is not None:
                head_uid = edge.parentRef.get("value")
                dot.edge(tail_uid, head_uid, edge_type)
            else:
                dot.edge(tail_uid, tail_uid, edge_type)
        return dot



#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Commander
# Description: 
#
#   This script demonstrates how to script the onboarding user accounts in an enterprise
#   and pushing records to the user's vault. 
#
#   - If the email domain has been reserved, the account is created instantly 
#     and the records are pushed.
#
#   - If the email domain is not reserved (e.g. gmail.com), the user will receive
#   an invite to create their vault. Upon next execution 
#
# Usage:
#    - Example: Invite a specific user:
#    python3 user_create.py <email>
#
#    - Example: Check the status of users and push records:
#    python3 user_create.py
#

import argparse
import getpass
import logging
import os
import re
import sys
from typing import Optional

from keepercommander import api
from keepercommander import vault, vault_extensions, record_management
from keepercommander.__main__ import get_params_from_config
from keepercommander.commands.enterprise import EnterpriseUserCommand
from keepercommander.commands.enterprise_create_user import CreateEnterpriseUserCommand
from keepercommander.commands.enterprise_push import EnterprisePushCommand

parser = argparse.ArgumentParser(description='Onboard users')
parser.add_argument('--debug', action='store_true', help='Enables debug logging')
parser.add_argument('email', nargs='?', help='User to create')
opts, flags = parser.parse_known_args(sys.argv[1:])
logging.basicConfig(level=logging.DEBUG if opts.debug is True else logging.WARNING, format='%(message)s')

# Enter the name of the node where you are onboarding the users
NODE_NAME = '<NODE NAME HERE>'

authKeeper = get_params_from_config(os.path.join(os.path.dirname(__file__), 'config.json'))

# Prompt for login unless the config exists
while not authKeeper.user:
    authKeeper.user = getpass.getpass(prompt='User(Email): ', stream=None)

def create_or_invite_user(params, email):
    try:
        # Attempt to create the account without an invitation
        create_command = CreateEnterpriseUserCommand()
        create_command.execute(params, email=email, node=NODE_NAME)
        # Get the latest enterprise data
        api.query_enterprise(authKeeper)
    except Exception as e:
        logging.warning(e)

    user = next((x for x in authKeeper.enterprise['users'] if x['username'] == email), None)
    if user is None:
        try:
            # User must be invited normally
            user_command = EnterpriseUserCommand()
            user_command.execute(params, invite=True, node=NODE_NAME, email=[email])
            api.query_enterprise(authKeeper)
        except Exception as e:
            logging.warning(e)

def push_records_to_user(params, email):
    script_dir = os.path.dirname(__file__)
    script_name = os.path.join(script_dir, 'enterprise_push.json')
    if os.path.isfile(script_name):
        command = EnterprisePushCommand()
        command.execute(params, user=[email], file=script_name)
    else:
        logging.error('Unable to locate JSON push file "%s"', script_name)

api.login(authKeeper)

if not authKeeper.session_token:
    logging.error('Unable to login to Keeper')
    exit(1)

api.query_enterprise(authKeeper)
if not authKeeper.enterprise:
    logging.error('Unable to retrieve enterprise data - not an enterprise administrator')
    exit(1)

username = opts.email
if username:
    username = opts.email.lower()

    # See if the user exists in the enterprise
    user = next((x for x in authKeeper.enterprise['users'] if x['username'] == username), None)
    if user is None:   # user does not exist
        if re.search(r'[\w.]+@[\w.]+', username):
            create_or_invite_user(authKeeper, username)
        else:
            logging.warning('Not a valid email address: %s', username)
            exit(1)

        # Check to see if we succeeded in creating or inviting the user
        user = next((x for x in authKeeper.enterprise['users'] if x['username'] == username), None)
        if user is None:
            logging.error(f'User {username} was not created/invited')
            exit(1)

# Search for a record in the vault with the name "Created users"
# and use this record for storing the status of invited users
api.sync_down(params=authKeeper)
record = next(vault_extensions.find_records(authKeeper, 'Created users'), None)   # type: Optional[vault.TypedRecord]
if record is None:
    # record not found, so create it
    record = vault.KeeperRecord.create(authKeeper, 'encryptedNotes')
    record.title = 'Created users'
    record_management.add_record_to_folder(authKeeper, record)
    api.sync_down(params=authKeeper)
    record = next(vault_extensions.find_records(authKeeper, 'Created users'), None)
    if record is None:
        logging.error(f'Record "Created users" cannot be created')
        exit(1)
    else:
        logging.warning(f'Record "Created users" created')

pushed_to = []
should_update_record = False
try:
    if username:
        field = next((x for x in record.custom if x.type == 'text' and x.label == username), None)
        if field is None:
            field = vault.TypedField.new_field('text', ['push'], username)
            record.custom.append(field)
            should_update_record = True

    for status_field in record.custom:
        if status_field.type != 'text':
            continue
        email = status_field.label
        if not email:
            continue
        status_value = status_field.get_default_value(str)
        if not status_value:
            continue
        if status_value.casefold() != 'push':
            continue

        # find user
        user = next((x for x in authKeeper.enterprise['users'] if x['username'] == email), None)
        if user is None:
            continue
        # Check the status of the user (it can be active or invited)
        status = user.get('status')
        if status != 'active':
            continue
        lock = user.get('lock')
        if isinstance(lock, int) and lock != 0:
            logging.warning('User "%s" is locked. Skipping.', email)
            continue

        logging.warning('Pushing records to user "%s"', email)
        try:
            push_records_to_user(authKeeper, email)
            status_field.value = ['done']
            should_update_record = True
            pushed_to.append(email)
        except Exception as e:
            logging.warning('Error pushing records to %s: %s', email, e)
finally:
    if should_update_record:
        if len(pushed_to) > 0:
            logging.warning('Pushed records to:')
            for email in pushed_to:
                no = 1
                logging.warning(f'{no:>4}. {email}')
                no += 1
        record_management.update_record(authKeeper, record)
